
#include "CudaLBFGS/lbfgs.h"
#include "CudaLBFGS/error_checking.h"
#include "CudaLBFGS/timer.h"

#include <iostream>
#include <vector>
#include <string>

#include <clipper/core/coords.h>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_functions.h>

#include "cuda_restraint.cuh"

#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n",   \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)


#ifdef __CUDACC__


// should this be a member function of cuda_restraint_t?
__device__
float get_distortion(const float *d_x, const cuda_restraint_t &rest) {

   // pre-calculate this?
   float weight =  1.0/(rest.sigma * rest.sigma);

   unsigned int idx_at_1 = 3 * rest.atom_idx_1;
   unsigned int idx_at_2 = 3 * rest.atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   float bl = sqrt(bl_sq);
   float delta =  bl - rest.bl_target;
   float distortion = weight * delta * delta;
   return distortion;
}


// fill the values of d_grad
__device__
void
use_numerical_gradients(const float *d_x, float *d_grad, const float &bl_target) {

   float x_modifiable[6];
   for (unsigned int i=0; i<6; i++)
      x_modifiable[i] = d_x[i];

   for (unsigned int i=0; i<6; i++)
      d_grad[i] = 0.0;

   // Iterate through restraints

   cuda_restraint_t rest;

   float weight = 3.0;
   float micro_step = 0.01;
   for (unsigned int i=0; i<6; i++) {
      float x_orig = d_x[i];
      x_modifiable[i] -= micro_step;
      float dist_1 = get_distortion(x_modifiable, rest);
      x_modifiable[i] = x_orig;
      x_modifiable[i] += micro_step;
      float dist_2 = get_distortion(x_modifiable, rest);
      float g = (dist_2 - dist_1)/(2.0 * micro_step); 
      x_modifiable[i] = x_orig;
      printf("gradient compare for %d : %f vs %f \n", i, d_grad[i], g);
      d_grad[i] += g;
   }

}

__device__
void add_gradients(const cuda_restraint_t &rest, const float *d_x, float *d_grad) {

   unsigned int idx_at_1 = 3 * rest.atom_idx_1;
   unsigned int idx_at_2 = 3 * rest.atom_idx_2;

   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   float bl = sqrt(bl_sq);
   float delta =  bl - rest.bl_target;
   float weight = 1.0/(rest.sigma * rest.sigma);

   float constant_part = 2.0 * weight * (bl - rest.bl_target)/bl;

   float x_k_contrib = constant_part * xd;
   float y_k_contrib = constant_part * yd;
   float z_k_contrib = constant_part * zd;

   float x_l_contrib = constant_part * -xd;
   float y_l_contrib = constant_part * -yd;
   float z_l_contrib = constant_part * -zd;

    // printf("3*atom_1_idx %d\n", 3*atom_1_idx);
    d_grad[idx_at_1  ] += x_k_contrib;
    d_grad[idx_at_1+1] += y_k_contrib;
    d_grad[idx_at_1+2] += z_k_contrib;

    // printf("3*atom_2_idx %d\n", 3*atom_2_idx);
    d_grad[idx_at_2  ] += x_l_contrib;
    d_grad[idx_at_2+1] += y_l_contrib;
    d_grad[idx_at_2+2] += z_l_contrib;

    // printf("d_grad[0] %f d_grad[1] %f d_grad[2] %f\n", d_grad[0], d_grad[1], d_grad[2]);
    // printf("d_grad[3] %f d_grad[4] %f d_grad[5] %f\n", d_grad[3], d_grad[4], d_grad[5]);


}


// pass: d_x: pointer to an array of "coordinates" (model),
//
// returns: d_f: target function value for these coordinates
//          d_grad: gradients for these coordinates
__global__
void kernel_combined_f_gradf(const float *d_x, float *d_y, float *d_grad) {

   unsigned int n_atoms = 2;

   // is there a faster way to zero the memory?
   for(unsigned int i=0; i<n_atoms*3; i++)
      d_grad[i] = 0.0;

   unsigned int bond_idx = 0;

   // Iterate through restraints

   cuda_restraint_t rest;

   float d_by_fn = get_distortion(d_x, rest);
   add_gradients(rest, d_x, d_grad);

   *d_y = d_by_fn;

   // use_numerical_gradients(d_x, d_grad, bl_target);
}

#endif // __CUDACC__



class gpu_bonds : public cost_function {

public:

   // the derived class must be constructed with an argument

   // We need to transfer the restraintss info here too
   // and transfer them to the GPU.
   gpu_bonds(unsigned int n_atoms) : cost_function(3*n_atoms) {}
   
   // pass: d_x: pointer to an array of "coordinates" (model),
   //
   // returns: d_f: target function value for these coordinates
   //          d_grad: gradients for these coordinates
   void f_gradf(const float *d_x, float *d_f, float *d_grad) {
          kernel_combined_f_gradf<<<1,1>>>(d_x, d_f, d_grad);
          cudaCheckErrors("cuda kernel fail kernel_combined_f_gradf()");
   }

};



int main(int argc, char **argv) {

   int status = 1;

   float *d_x;

   // atoms
   std::vector<clipper::Coord_orth> atom_positions;
   atom_positions.push_back(clipper::Coord_orth(1.5, 2.6, 3.1));
   atom_positions.push_back(clipper::Coord_orth(5.2, 6.3, 7.2));

   unsigned int n_atoms = atom_positions.size();

   gpu_bonds gb(n_atoms);
   lbfgs minimizer(gb);

   // allocate host memory:
   float *x = new float[n_atoms * 3];
   // allocate d_x on the device
   cudaMalloc(&d_x, n_atoms * 3 * sizeof(float));
   // now fill x
   for (unsigned int i=0; i<atom_positions.size(); i++) {
      x[3*i  ] = atom_positions[i].x();
      x[3*i+1] = atom_positions[i].y();
      x[3*i+2] = atom_positions[i].z();
      std::cout << "Input i: " << i << " x: " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;
   }
   // copy x to d_x
   cudaMemcpy(d_x, x, n_atoms * 3 * sizeof(float), cudaMemcpyHostToDevice);

   std::cout << "Minimize..." << std::endl;
   minimizer.minimize(d_x);
   cudaMemcpy(x, d_x, n_atoms * 3  * sizeof(float), cudaMemcpyDeviceToHost);
   for (unsigned int i=0; i<atom_positions.size(); i++) {
      std::cout << " Result " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;
   }
   float bl_x_2 = x[3] - x[0];
   float bl_y_2 = x[4] - x[1];
   float bl_z_2 = x[5] - x[2];
   float bl_2 = bl_x_2 * bl_x_2 + bl_y_2 * bl_y_2 + bl_z_2 * bl_z_2;
   float bl = sqrt(bl_2);
   std::cout << " Result bl " << bl << std::endl;

   cudaFree(d_x);
   delete [] x;

   return status;

}
