
// -*-c++-*-

#include <CudaLBFGS/error_checking.h>
#include <CudaLBFGS/timer.h>

#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <chrono>

#ifndef STANDALONE_INPUTS
#include <clipper/core/coords.h>
#endif

#define HAVE_GSL  // This is a hack to compile simple-restraint.hh
                  // we need a better mechanism

#ifndef STANDALONE_INPUTS
#include <coot/ideal/simple-restraint.hh>
#endif

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_functions.h>
#include <math_constants.h> // for pi

#include "gpu-bonds.hh"

// maps
#ifndef STANDALONE_INPUTS
#include "coot/coot-utils/atom-selection-container.hh"
#include "coot-utils/coot-coord-utils.hh"
#include "coot/coot-utils/coot-map-utils.hh"
#include "coot/coot-utils/coot-map-heavy.hh"

#include "clipper/ccp4/ccp4_map_io.h"
#endif // STANDALONE_INPUTS

#include "atoms-container-t.hh"
#include "cuda-minimize.hh"


#ifdef __CUDACC__

#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n",   \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)


float
cuda_restraints_container_t::distortion(float *d_x, bool do_bonds, bool do_angles, bool do_torsions, bool do_chiral_volumes, bool do_planes, bool do_nbcs,
                                        bool do_geman_mcclure_distances) const {

   float sum = 0.0;
   if (do_bonds) {
      for (unsigned int i=0; i<n_bond_restraints; i++) {
         const cuda_bond_restraint_t &rest = bonds[i];
         sum += rest.distortion(d_x);
      }
   }
   if (do_angles) {
      for (unsigned int i=0; i<n_angle_restraints; i++) {
         const cuda_angle_restraint_t &rest = angles[i];
         sum += rest.distortion(d_x);
      }
   }
   if (do_torsions) {
      for (unsigned int i=0; i<n_torsion_restraints; i++) {
         const cuda_torsion_restraint_t &rest = torsions[i];
         sum += rest.distortion(d_x);
      }
   }
   if (do_chiral_volumes) {
      for (unsigned int i=0; i<n_chiral_volume_restraints; i++) {
         const cuda_chiral_volume_restraint_t &rest = chiral_volumes[i];
         sum += rest.distortion(d_x);
      }
   }
   if (do_planes) {
      for (unsigned int i=0; i<n_plane_restraints; i++) {
         const cuda_device_plane_restraint_t &rest = planes[i];
         // printf("plane restraint number %d has distortion %f\n", i, rest.distortion(d_x));
         sum += rest.distortion(d_x);
      }
   }
   if (do_nbcs) {
      for (unsigned int i=0; i<n_nbc_restraints; i++) {
         const cuda_nbc_restraint_t &rest = nbcs[i];
         sum += rest.distortion(d_x);
      }
   }
   if (do_geman_mcclure_distances) {
      for (unsigned int i=0; i<n_geman_mcclure_restraints; i++) {
         const cuda_geman_mcclure_restraint_t &rest = geman_mcclures[i];
         // printf("GM restraint number %d has distortion %f\n", i, rest.distortion(d_x));
         sum += rest.distortion(d_x);
      }
   }
   return sum;
}


// For testing ther derivatives:
// I thihk that this should be its own kernel (const ref restraints needs fixing?) called with 1,1 launch parameters.
//
// should be part of cuda_restraints_container_t?
__device__
void
numerical_vs_analytical_gradients(const cuda_restraints_container_t &restraints, int n_atoms, float *d_x, float *d_grad) {

   float micro_step = 0.01;

   bool do_bonds = true;
   bool do_angles = false;
   bool do_torsions = false;
   bool do_chiral_volumes = false;
   bool do_planes = false;
   bool do_nbcs = false;
   bool do_geman_mcclure_distances = false;

   // test bonds
   for (unsigned int i=0; i<restraints.n_bond_restraints; i++) {
       const cuda_bond_restraint_t &rest = restraints.bonds[i];
       rest.add_gradients(d_x, d_grad);
   }

#if 0
   // To make torsion analytical restraints:
   // iterate through restraints.n_torsion_restraints

   for (unsigned int i=0; i<restraints.n_torsion_restraints; i++) {
       const cuda_torsion_restraint_t &rest = restraints.torsions[i];
       rest.add_gradients(d_x, d_grad);
   }
#endif

   double sum_distortion = 0.0;
   for (int idx=0; idx<3*n_atoms; idx++) {
      float grad_cur = d_grad[idx]; // analytical
      float pos_orig = d_x[idx];
      d_x[idx] -= micro_step;
      double dw1 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig + micro_step;
      double dw2 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig;

      double grad_numerical = (dw2-dw1)/(2.0 * micro_step);

      d_grad[idx] = grad_numerical;
      printf("numerical gradients %d was %14.5f now %14.5f\n", idx, grad_cur, grad_numerical);
   }
}

// fill the values of d_grad
__device__
void
use_numerical_gradients(const cuda_restraints_container_t &restraints, int n_atoms, float *d_x, float *d_grad,
                        bool do_bonds, bool do_angles, bool do_torsions, bool do_chiral_volumes, bool do_planes, bool do_nbcs,
                        bool do_geman_mcclure_distances) {

   float micro_step = 0.01;

   double sum_distortion = 0.0;
   for (int idx=0; idx<3*n_atoms; idx++) {
      float grad_cur = d_grad[idx];
      float pos_orig = d_x[idx];
      d_x[idx] -= micro_step;
      double dw1 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig + micro_step;
      double dw2 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig;

      do_torsions = false;

      d_x[idx] -= micro_step;
      double ds1 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig + micro_step;
      double ds2 = restraints.distortion(d_x, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);
      d_x[idx] = pos_orig;

      // isolate the torsions
      double grad = (dw2-dw1-ds1+ds2)/(2.0 * micro_step);

      d_grad[idx] = grad;
      printf("numerical gradients %d was %14.5f now %14.5f\n", idx, grad_cur, grad);
   }


}


__device__
float
cuda_bond_restraint_t::distortion(const float *d_x) const {

   // pre-calculate this?
   float weight = 1.0f/(sigma * sigma);

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;

   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   float bl = sqrt(bl_sq);
   float delta =  bl - target_value;
   float distortion = weight * delta * delta;
   // printf("bond distortion for atoms %d %d is %f\n", atom_idx_1, atom_idx_2, distortion);
   return distortion;
}



__device__
void
cuda_bond_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;

   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   // printf("in add_gradients bl_sq %f\n", bl_sq);

   float bl = sqrt(bl_sq);
   float delta =  bl - target_value;
   float weight = 1.0f/(sigma * sigma);

   float constant_part = 2.0 * weight * (bl - target_value)/bl;

   float x_k_contrib = constant_part * xd;
   float y_k_contrib = constant_part * yd;
   float z_k_contrib = constant_part * zd;

   float x_l_contrib = constant_part * -xd;
   float y_l_contrib = constant_part * -yd;
   float z_l_contrib = constant_part * -zd;

   if (false)
       printf("in bond::add_gradients() atom_idx_1 %d atom_idx_2 %d adding %10.3f %10.3f %10.3f\n",
              atom_idx_1, atom_idx_2, x_k_contrib, y_k_contrib, z_k_contrib);

    atomicAdd(d_grad+idx_at_1  , x_k_contrib);
    atomicAdd(d_grad+idx_at_1+1, y_k_contrib);
    atomicAdd(d_grad+idx_at_1+2, z_k_contrib);

    atomicAdd(d_grad+idx_at_2  , x_l_contrib);
    atomicAdd(d_grad+idx_at_2+1, y_l_contrib);
    atomicAdd(d_grad+idx_at_2+2, z_l_contrib);

}

void
cuda_bond_restraint_t::get_gradients(const float *d_x, float *d_grad, float *local_gradients) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;

   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   // printf("in add_gradients bl_sq %f\n", bl_sq);

   float bl = sqrt(bl_sq);
   float delta =  bl - target_value;
   float weight = 1.0/(sigma * sigma);

   float constant_part = 2.0 * weight * (bl - target_value)/bl;

   float x_k_contrib = constant_part * xd;
   float y_k_contrib = constant_part * yd;
   float z_k_contrib = constant_part * zd;

   float x_l_contrib = constant_part * -xd;
   float y_l_contrib = constant_part * -yd;
   float z_l_contrib = constant_part * -zd;

   if (false)
       printf("in bond::add_gradients() atom_idx_1 %d atom_idx_2 %d adding %10.3f %10.3f %10.3f\n",
              atom_idx_1, atom_idx_2, x_k_contrib, y_k_contrib, z_k_contrib);

    // atomicAdd(d_grad+idx_at_1  , x_k_contrib);
    // atomicAdd(d_grad+idx_at_1+1, y_k_contrib);
    // atomicAdd(d_grad+idx_at_1+2, z_k_contrib);

    // atomicAdd(d_grad+idx_at_2  , x_l_contrib);
    // atomicAdd(d_grad+idx_at_2+1, y_l_contrib);
    // atomicAdd(d_grad+idx_at_2+2, z_l_contrib);

    local_gradients[idx_at_1  ] += x_k_contrib;
    local_gradients[idx_at_1+1] += y_k_contrib;
    local_gradients[idx_at_1+2] += z_k_contrib;

    local_gradients[idx_at_2  ] += x_l_contrib;
    local_gradients[idx_at_2+1] += y_l_contrib;
    local_gradients[idx_at_2+2] += z_l_contrib;


}



__device__
void
cuda_angle_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;

   // use float3s?
   float xd_len1 = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd_len1 = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd_len1 = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float xd_len2 = d_x[idx_at_3  ] - d_x[idx_at_2  ];
   float yd_len2 = d_x[idx_at_3+1] - d_x[idx_at_2+1];
   float zd_len2 = d_x[idx_at_3+2] - d_x[idx_at_2+2];

   float len1_sqd = xd_len1*xd_len1 + yd_len1*yd_len1 + zd_len1*zd_len1;
   float len2_sqd = xd_len2*xd_len2 + yd_len2*yd_len2 + zd_len2*zd_len2;

   float a = sqrt(len1_sqd);
   float b = sqrt(len2_sqd);

   if (a < 0.01) { a = 0.01f; }
   if (b < 0.01) { b = 0.01f; }

   float l_over_a_sqd = 1.0f/(a*a);
   float l_over_b_sqd = 1.0f/(b*b);
   float l_ab         = 1.0f/(a*b);

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);
   float a_dot_b = cuda_lau::dot(d1, d2);

   float cos_theta = a_dot_b/(a*b);
   if (cos_theta < -1) cos_theta = -1.0f;
   if (cos_theta >  1) cos_theta =  1.0f;
   float theta = acos(cos_theta);
   if (theta < 0.001f) theta = 0.001f; // never happens?

   float prem = -1.0f/sin(theta);

   // k: position of atom_1
   // l: position of atom_2
   // m: position of atom_3
   const float k[3] = {d_x[idx_at_1], d_x[idx_at_1+1], d_x[idx_at_1+2]};
   const float l[3] = {d_x[idx_at_2], d_x[idx_at_2+1], d_x[idx_at_2+2]};
   const float m[3] = {d_x[idx_at_3], d_x[idx_at_3+1], d_x[idx_at_3+2]};

   // printf("prem %f cos_theta %f l_over_a_sqd %14e \n", prem, cos_theta, l_over_a_sqd);
   float x_k_contrib = prem*(cos_theta*(l[0]-k[0])*l_over_a_sqd + l_ab*(m[0]-l[0]));
   float y_k_contrib = prem*(cos_theta*(l[1]-k[1])*l_over_a_sqd + l_ab*(m[1]-l[1]));
   float z_k_contrib = prem*(cos_theta*(l[2]-k[2])*l_over_a_sqd + l_ab*(m[2]-l[2]));

   float x_m_contrib = prem*(cos_theta*(l[0]-m[0])*l_over_b_sqd + l_ab*(k[0]-l[0]));
   float y_m_contrib = prem*(cos_theta*(l[1]-m[1])*l_over_b_sqd + l_ab*(k[1]-l[1]));
   float z_m_contrib = prem*(cos_theta*(l[2]-m[2])*l_over_b_sqd + l_ab*(k[2]-l[2]));

   // middle atom

   float term1x = (-cos_theta*(l[0]-k[0])*l_over_a_sqd) - cos_theta*(l[0]-m[0])*l_over_b_sqd;
   float term1y = (-cos_theta*(l[1]-k[1])*l_over_a_sqd) - cos_theta*(l[1]-m[1])*l_over_b_sqd;
   float term1z = (-cos_theta*(l[2]-k[2])*l_over_a_sqd) - cos_theta*(l[2]-m[2])*l_over_b_sqd;

   float term2x = (-(k[0]-l[0]) - (m[0]-l[0])) * l_ab;
   float term2y = (-(k[1]-l[1]) - (m[1]-l[1])) * l_ab;
   float term2z = (-(k[2]-l[2]) - (m[2]-l[2])) * l_ab;

   float x_l_mid_contrib = prem * (term1x + term2x);
   float y_l_mid_contrib = prem * (term1y + term2y);
   float z_l_mid_contrib = prem * (term1z + term2z);

   float weight = 1.0f/(sigma * sigma);
   float rad_to_deg = 180.0f/M_PI;
   float deg_to_rad = M_PI/180.0;
   float ds_dth = 2.0f * (rad_to_deg * theta - target_value);
   float w_ds_dth = weight * ds_dth * rad_to_deg;

   {
      if (false) {
          printf("in angle::add_gradients() A atom_idx_1 %d atom_idx_2 %d atom_idx_3 %d adding %14e %14e %14e and w_ds_dth %f\n",
                 atom_idx_1, atom_idx_2, atom_idx_3, x_k_contrib, y_k_contrib, z_k_contrib, w_ds_dth);
          printf("in angle::add_gradients() B atom_idx_1 %d atom_idx_2 %d atom_idx_3 %d adding %14e %14e %14e and w_ds_dth %f\n",
                 atom_idx_1, atom_idx_2, atom_idx_3, x_m_contrib, y_m_contrib, z_m_contrib, w_ds_dth);
          printf("in angle::add_gradients() C atom_idx_1 %d atom_idx_2 %d atom_idx_3 %d adding %14e %14e %14e and w_ds_dth %f\n",
                 atom_idx_1, atom_idx_2, atom_idx_3, x_l_mid_contrib, y_l_mid_contrib, z_l_mid_contrib, w_ds_dth);
      }

      atomicAdd(d_grad+idx_at_1  , x_k_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_1+1, y_k_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_1+2, z_k_contrib*w_ds_dth);
   }
   {
      atomicAdd(d_grad+idx_at_3  , x_m_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_3+1, y_m_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_3+2, z_m_contrib*w_ds_dth);
   }
   {
      atomicAdd(d_grad+idx_at_2  , x_l_mid_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_2+1, y_l_mid_contrib*w_ds_dth);
      atomicAdd(d_grad+idx_at_2+2, z_l_mid_contrib*w_ds_dth);
   }
}

__device__
torsion_distortion_gradients_t::torsion_distortion_gradients_t(const int &P1_idx,
                                                               const int &P2_idx,
                                                               const int &P3_idx,
                                                               const int &P4_idx,
                                                               const float *d_x) {

   unsigned int idx_at_1 = 3 * P1_idx;
   unsigned int idx_at_2 = 3 * P2_idx;
   unsigned int idx_at_3 = 3 * P3_idx;
   unsigned int idx_at_4 = 3 * P4_idx;

   float3 P1 = make_float3(d_x[idx_at_1], d_x[idx_at_1+1], d_x[idx_at_1+2]);
   float3 P2 = make_float3(d_x[idx_at_3], d_x[idx_at_2+1], d_x[idx_at_2+2]);
   float3 P3 = make_float3(d_x[idx_at_3], d_x[idx_at_3+1], d_x[idx_at_3+2]);
   float3 P4 = make_float3(d_x[idx_at_4], d_x[idx_at_4+1], d_x[idx_at_4+2]);

   float3 a = make_float3(P2.x-P1.x, P2.y-P1.y, P2.z-P1.z);
   float3 b = make_float3(P3.x-P2.x, P3.y-P2.y, P3.z-P2.z);
   float3 c = make_float3(P4.x-P3.x, P4.y-P3.y, P4.z-P3.z);

   float a_lengthsq = a.x*a.x + a.y*a.y + a.z*a.z;
   float a_length = sqrt(a_lengthsq);
   float b_lengthsq = b.x*b.x + b.y*b.y + b.z*b.z;
   float b_length = sqrt(b_lengthsq);
   float c_lengthsq = c.x*c.x + c.y*c.y + c.z*c.z;
   float c_length = sqrt(c_lengthsq);

   float H = -cuda_lau::dot(a,c);
   float J =  cuda_lau::dot(a,b); 
   float K =  cuda_lau::dot(b,c); 
   float L = 1.0f/b_lengthsq;
   float one_over_b = 1.0f/b_length;

   // Is this block correct?

   // a.(bxc)/b
   float3 cp;
   cuda_lau::cross(b,c, cp); 
   float E = one_over_b * cuda_lau::dot(a, cp);
   // -a.c+(a.b)(b.c)/(b*b)
   float G = H+J*K*L;
   float F = 1.0f/G;

   tan_phi = E/G;
   phi = atan2(E,G); // radians

   float al = a_length;
   float cl = c_length;
   float cos_a1 = cuda_lau::dot(a,b)/(al*b_length);
   float cos_a2 = cuda_lau::dot(b,c)/(b_length*cl);

   // x
   float dH_dxP1 =  c.x;
   float dH_dxP2 = -c.x;
   float dH_dxP3 =  a.x;
   float dH_dxP4 = -a.x;

   float dK_dxP1 = 0; 
   float dK_dxP2 = -c.x;
   float dK_dxP3 =  c.x - b.x;
   float dK_dxP4 =  b.x;

   float dJ_dxP1 = -b.x;
   float dJ_dxP2 =  b.x - a.x;
   float dJ_dxP3 =  a.x;
   float dJ_dxP4 =  0;

   float dL_dxP1 = 0;
   float dL_dxP2 =  2*(P3.x-P2.x)*L*L;
   float dL_dxP3 = -2*(P3.x-P2.x)*L*L;
   float dL_dxP4 = 0;

   // y 
   float dH_dyP1 =  c.y;
   float dH_dyP2 = -c.y;
   float dH_dyP3 =  a.y;
   float dH_dyP4 = -a.y;

   float dK_dyP1 = 0; 
   float dK_dyP2 = -c.y;
   float dK_dyP3 =  c.y - b.y;
   float dK_dyP4 =  b.y;

   float dJ_dyP1 = -b.y;
   float dJ_dyP2 =  b.y - a.y;
   float dJ_dyP3 =  a.y;
   float dJ_dyP4 =  0;

   float dL_dyP1 = 0;
   float dL_dyP2 =  2*(P3.y - P2.y)*L*L; // check sign
   float dL_dyP3 = -2*(P3.y - P2.y)*L*L;
   float dL_dyP4 = 0;

   // z 
   float dH_dzP1 =  c.z;
   float dH_dzP2 = -c.z;
   float dH_dzP3 =  a.z;
   float dH_dzP4 = -a.z;

   float dK_dzP1 = 0; 
   float dK_dzP2 = -c.z;
   float dK_dzP3 =  c.z - b.z;
   float dK_dzP4 =  b.z;

   float dJ_dzP1 = -b.z;
   float dJ_dzP2 =  b.z - a.z;
   float dJ_dzP3 =  a.z;
   float dJ_dzP4 =  0;

   float dL_dzP1 = 0;
   float dL_dzP2 =  2*(P3.z-P2.z)*L*L;
   float dL_dzP3 = -2*(P3.z-P2.z)*L*L;
   float dL_dzP4 = 0;

   // M
   float dM_dxP1 = -(b.y*c.z - b.z*c.y);
   float dM_dxP2 =  (b.y*c.z - b.z*c.y) + (a.y*c.z - a.z*c.y);
   float dM_dxP3 =  (b.y*a.z - b.z*a.y) - (a.y*c.z - a.z*c.y);
   float dM_dxP4 = -(b.y*a.z - b.z*a.y);

   float dM_dyP1 = -(b.z*c.x - b.x*c.z);
   float dM_dyP2 =  (b.z*c.x - b.x*c.z) + (a.z*c.x - a.x*c.z);
   float dM_dyP3 = -(a.z*c.x - a.x*c.z) + (b.z*a.x - b.x*a.z);
   float dM_dyP4 = -(b.z*a.x - b.x*a.z);

   float dM_dzP1 = -(b.x*c.y - b.y*c.x);
   float dM_dzP2 =  (b.x*c.y - b.y*c.x) + (a.x*c.y - a.y*c.x);
   float dM_dzP3 = -(a.x*c.y - a.y*c.x) + (a.y*b.x - a.x*b.y);
   float dM_dzP4 = -(a.y*b.x - a.x*b.y);

   // E
   float dE_dxP1 = dM_dxP1*one_over_b;
   float dE_dyP1 = dM_dyP1*one_over_b;
   float dE_dzP1 = dM_dzP1*one_over_b; 

   // M = Eb
   float dE_dxP2 = dM_dxP2*one_over_b + E*(P3.x - P2.x)*L;
   float dE_dyP2 = dM_dyP2*one_over_b + E*(P3.y - P2.y)*L;
   float dE_dzP2 = dM_dzP2*one_over_b + E*(P3.z - P2.z)*L;
    
   float dE_dxP3 = dM_dxP3*one_over_b - E*(P3.x - P2.x)*L;
   float dE_dyP3 = dM_dyP3*one_over_b - E*(P3.y - P2.y)*L;
   float dE_dzP3 = dM_dzP3*one_over_b - E*(P3.z - P2.z)*L;
    
   float dE_dxP4 = dM_dxP4*one_over_b;
   float dE_dyP4 = dM_dyP4*one_over_b;
   float dE_dzP4 = dM_dzP4*one_over_b;

   float EFF = E*F*F;
   float JL = J*L;
   float KL = K*L;
   float JK = J*K;

   // x
   dD_dxP1 = F*dE_dxP1 - EFF*(dH_dxP1 + JL*dK_dxP1 + KL*dJ_dxP1 + JK*dL_dxP1);
   dD_dxP2 = F*dE_dxP2 - EFF*(dH_dxP2 + JL*dK_dxP2 + KL*dJ_dxP2 + JK*dL_dxP2);
   dD_dxP3 = F*dE_dxP3 - EFF*(dH_dxP3 + JL*dK_dxP3 + KL*dJ_dxP3 + JK*dL_dxP3);
   dD_dxP4 = F*dE_dxP4 - EFF*(dH_dxP4 + JL*dK_dxP4 + KL*dJ_dxP4 + JK*dL_dxP4);

   // y
   dD_dyP1 = F*dE_dyP1 - EFF*(dH_dyP1 + JL*dK_dyP1 + KL*dJ_dyP1 + JK*dL_dyP1);
   dD_dyP2 = F*dE_dyP2 - EFF*(dH_dyP2 + JL*dK_dyP2 + KL*dJ_dyP2 + JK*dL_dyP2);
   dD_dyP3 = F*dE_dyP3 - EFF*(dH_dyP3 + JL*dK_dyP3 + KL*dJ_dyP3 + JK*dL_dyP3);
   dD_dyP4 = F*dE_dyP4 - EFF*(dH_dyP4 + JL*dK_dyP4 + KL*dJ_dyP4 + JK*dL_dyP4);

   // z
   dD_dzP1 = F*dE_dzP1 - EFF*(dH_dzP1 + JL*dK_dzP1 + KL*dJ_dzP1 + JK*dL_dzP1);
   dD_dzP2 = F*dE_dzP2 - EFF*(dH_dzP2 + JL*dK_dzP2 + KL*dJ_dzP2 + JK*dL_dzP2);
   dD_dzP3 = F*dE_dzP3 - EFF*(dH_dzP3 + JL*dK_dzP3 + KL*dJ_dzP3 + JK*dL_dzP3);
   dD_dzP4 = F*dE_dzP4 - EFF*(dH_dzP4 + JL*dK_dzP4 + KL*dJ_dzP4 + JK*dL_dzP4);

}


__device__
void
cuda_torsion_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;
   unsigned int idx_at_4 = 3 * atom_idx_4;
   
   float xd_len1 = d_x[idx_at_2  ] - d_x[idx_at_1  ];
   float yd_len1 = d_x[idx_at_2+1] - d_x[idx_at_1+1];
   float zd_len1 = d_x[idx_at_2+2] - d_x[idx_at_1+2];

   float xd_len2 = d_x[idx_at_3  ] - d_x[idx_at_2  ];
   float yd_len2 = d_x[idx_at_3+1] - d_x[idx_at_2+1];
   float zd_len2 = d_x[idx_at_3+2] - d_x[idx_at_2+2];

   float xd_len3 = d_x[idx_at_4  ] - d_x[idx_at_3  ];
   float yd_len3 = d_x[idx_at_4+1] - d_x[idx_at_3+1];
   float zd_len3 = d_x[idx_at_4+2] - d_x[idx_at_3+2];

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);
   float3 d3 = make_float3(xd_len3, yd_len3, zd_len3);

   float len2_sqd = xd_len2*xd_len2 + yd_len2*yd_len2 + zd_len2*zd_len2;
   float len2 = sqrt(len2_sqd);

   float3 cp;
   cuda_lau::cross(d2, d3, cp);
   float E = cuda_lau::dot(d1, cp) * len2;
   float G = -cuda_lau::dot(d1, d3)*len2_sqd + cuda_lau::dot(d1, d2) * cuda_lau::dot(d2, d3);
   float phi = atan2(E, G); // radians

   float phi_0_jk_rad = (CUDART_PI_F/180.0) * phi0_jk;

   // float distortion = 0.5 * V_jk * (1.0f - cos(n_jk*(phi - phi_0_jk_rad)));
   //
   float dV_dphi = 0.5 * V_jk * (sin(n_jk*(phi - phi0_jk))) * n_jk;

   // dV/dx = dV/dphi * dphi/x

   torsion_distortion_gradients_t tdg(atom_idx_1, atom_idx_2, atom_idx_3, atom_idx_4, d_x);

   float tt = tdg.tan_phi;
   float scale = dV_dphi/(1.0f + tt*tt);

   atomicAdd(d_grad+idx_at_1  , scale*tdg.dD_dxP1);
   atomicAdd(d_grad+idx_at_1+1, scale*tdg.dD_dyP1);
   atomicAdd(d_grad+idx_at_1+2, scale*tdg.dD_dzP1);

   atomicAdd(d_grad+idx_at_2  , scale*tdg.dD_dxP2);
   atomicAdd(d_grad+idx_at_2+1, scale*tdg.dD_dyP2);
   atomicAdd(d_grad+idx_at_2+2, scale*tdg.dD_dzP2);

   atomicAdd(d_grad+idx_at_3  , scale*tdg.dD_dxP3);
   atomicAdd(d_grad+idx_at_3+1, scale*tdg.dD_dyP3);
   atomicAdd(d_grad+idx_at_3+2, scale*tdg.dD_dzP3);

   atomicAdd(d_grad+idx_at_4  , scale*tdg.dD_dxP4);
   atomicAdd(d_grad+idx_at_4+1, scale*tdg.dD_dyP4);
   atomicAdd(d_grad+idx_at_4+2, scale*tdg.dD_dzP4);

}


__device__
float
cuda_nbc_restraint_t::distortion(const float *d_x) const {

   // perhaps we should pass this? or build it into the class?
   // float lj_epsilon = 0.3; // now in class

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float dist_sq = xd*xd + yd*yd + zd*zd;

   // float bl = sqrt(bl_sq);
   // float delta =  bl - target_value;

   float V_lj = 0;

   // the value lj_sigma is r when is the potential is 0.
   // for lj_sigma + delta the potential is negative
   // for lj_sigma - delta the potential is positive
   // the potential is at a minimum at lj_r_min
   // 
   // so if target_value is say 3.4A, lj_sigma is 3.4 sigma
   // and lj_r_min ~ 3.4 * 1.122 = 3.82
   float lj_sigma = target_value;
   // double lj_r_min = std::pow(2.0, 1.0/6.0) * lj_sigma;
   float lj_r_min = 1.122462048309373 * lj_sigma;

   float max_dist = 2.5 * lj_sigma; // r_max

   if (dist_sq < 0.81) dist_sq = 0.81; // 0.9^2
   float alpha_sqrd = lj_r_min*lj_r_min/dist_sq;
   float alpha_up_6  = alpha_sqrd * alpha_sqrd * alpha_sqrd;
   float alpha_up_12 = alpha_up_6 * alpha_up_6;
   V_lj = lj_epsilon * (alpha_up_12 - 2.0 * alpha_up_6);

   // offset the Vlj so that it is zero at r_max (beyond which we no longer
   // consider contributions to the distortion)

   float Vlj_at_rmax = -0.016316891136 * lj_epsilon; // see Lennard-Jones truncated and shifted for

   V_lj += Vlj_at_rmax;

   // printf("nbc distortion V_lj %d %d is %f\n", atom_idx_1, atom_idx_2, V_lj);
   return V_lj;
}


__device__
float
cuda_nbc_restraint_t::distortion_harmonic_approximation(const float *d_x) const {


   const float distance_sqrd_to_distortion_scale = 1.0/(sigma*sigma); // square of 1/0.04;
   float tvsqrd = target_value * target_value;
   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float dist_sq = xd*xd + yd*yd + zd*zd;

   float distortion = 0.0;
   
   if (dist_sq < tvsqrd) {
      float dist = sqrt(dist_sq);
      float delta = dist - target_value;
      distortion = distance_sqrd_to_distortion_scale * delta * delta;
   }
   
   return distortion;

}

__device__
void
cuda_nbc_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   // perhaps we should pass this?
   // float lj_epsilon = 0.3; // now in class

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float dist_sq = xd*xd + yd*yd + zd*zd;

   // float bl = sqrt(bl_sq);
   // float delta =  bl - target_value;

   float V_lj = 0;

   // the value lj_sigma is r when is the potential is 0.
   // for lj_sigma + delta the potential is negative
   // for lj_sigma - delta the potential is positive
   // the potential is at a minimum at lj_r_min
   // 
   // so if target_value is say 3.4A, lj_sigma is 3.4 sigma
   // and lj_r_min ~ 3.4 * 1.122 = 3.82
   float lj_sigma = target_value;
   // double lj_r_min = std::pow(2.0, 1.0/6.0) * lj_sigma;
   float lj_r_min = 1.122462048309373 * lj_sigma;

   float max_dist = 2.5 * lj_sigma; // r_max

   if (dist_sq < 0.81) dist_sq = 0.81; // 0.9^2

   if (dist_sq < (max_dist * max_dist)) {

      // double lj_r_min = pow(2.0, 1.0/6.0) * lj_sigma; // precalculate this pow value - done
      float lj_r_min = 1.122462048309373 * lj_sigma;
      float lj_r = sqrt(dist_sq);
      float alpha = lj_r_min/lj_r;
      float dalpha_dr = -lj_r_min/dist_sq;

      float alpha_sqrd = lj_r_min*lj_r_min/dist_sq;
      float alpha_up_5  = alpha_sqrd * alpha_sqrd * alpha;
      float alpha_up_6  = alpha_sqrd * alpha_sqrd * alpha_sqrd;
      float alpha_up_11 = alpha_up_6 * alpha_up_5;
      // double dVlj_dalpha = 12.0 * lj_epsilon * (std::pow(alpha, 11) - std::pow(alpha, 5));
      float dVlj_dalpha = 12.0 * lj_epsilon * (alpha_up_11 - alpha_up_5);

      float dVlj_dr = dVlj_dalpha * dalpha_dr;

      float constant_part = dVlj_dr/lj_r; // this is why we need the square root

      float shift_lim = 10000.1; // needs investigation

      { // test for fixed
         float x_k_contrib =  constant_part*xd;
         float y_k_contrib =  constant_part*yd;
         float z_k_contrib =  constant_part*zd;

         if (x_k_contrib >  shift_lim) x_k_contrib =  shift_lim;
         if (y_k_contrib >  shift_lim) y_k_contrib =  shift_lim;
         if (z_k_contrib >  shift_lim) z_k_contrib =  shift_lim;
         if (x_k_contrib < -shift_lim) x_k_contrib = -shift_lim;
         if (y_k_contrib < -shift_lim) y_k_contrib = -shift_lim;
         if (z_k_contrib < -shift_lim) z_k_contrib = -shift_lim;

         atomicAdd(d_grad+idx_at_1  , x_k_contrib);
         atomicAdd(d_grad+idx_at_1+1, y_k_contrib);
         atomicAdd(d_grad+idx_at_1+2, z_k_contrib);
         
         // printf("in nbc::add_gradients() atom_idx_1 %d atom_idx_2 %d adding %10.3f %10.3f %10.3f\n",
         // atom_idx_1, atom_idx_2, x_k_contrib, y_k_contrib, z_k_contrib);

      }

      {
         float x_l_contrib = -constant_part*xd;
         float y_l_contrib = -constant_part*yd;
         float z_l_contrib = -constant_part*zd;

         if (x_l_contrib >  shift_lim) x_l_contrib =  shift_lim;
         if (y_l_contrib >  shift_lim) y_l_contrib =  shift_lim;
         if (z_l_contrib >  shift_lim) z_l_contrib =  shift_lim;
         if (x_l_contrib < -shift_lim) x_l_contrib = -shift_lim;
         if (y_l_contrib < -shift_lim) y_l_contrib = -shift_lim;
         if (z_l_contrib < -shift_lim) z_l_contrib = -shift_lim;

         atomicAdd(d_grad+idx_at_2  , x_l_contrib);
         atomicAdd(d_grad+idx_at_2+1, y_l_contrib);
         atomicAdd(d_grad+idx_at_2+2, z_l_contrib);
      }
   }
}

__device__
void
cuda_nbc_restraint_t::add_gradients_harmonic_approximation(const float *d_x, float *d_grad) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float dist_sq = xd*xd + yd*yd + zd*zd;

   if (dist_sq < target_value * target_value) {
      float dist = sqrt(dist_sq);
      float weight = 1.0/(sigma * sigma);
      // do we have fast inverse sqrt on cuda?
      // float constant_part = 2.0*weight * (1.0 - target_val * f_inv_fsqrt(b_i_sqrd));
      float constant_part = 2.0*weight * (1.0 - target_value / dist);

      { // test for fixed
         float x_k_contrib =  constant_part*xd;
         float y_k_contrib =  constant_part*yd;
         float z_k_contrib =  constant_part*zd;

         atomicAdd(d_grad+idx_at_1  , x_k_contrib);
         atomicAdd(d_grad+idx_at_1+1, y_k_contrib);
         atomicAdd(d_grad+idx_at_1+2, z_k_contrib);
      }

      {
         float x_l_contrib = -constant_part*xd;
         float y_l_contrib = -constant_part*yd;
         float z_l_contrib = -constant_part*zd;

         atomicAdd(d_grad+idx_at_2  , x_l_contrib);
         atomicAdd(d_grad+idx_at_2+1, y_l_contrib);
         atomicAdd(d_grad+idx_at_2+2, z_l_contrib);
      }
   }
}


__host__ __device__
float
cuda_lau::dot(const float3 &d1, const float3 &d2) {

   return d1.x*d2.x + d1.y*d2.y+ d1.z*d2.z;

}

__host__ __device__
void
cuda_lau::cross(const float3 &d1, const float3 &d2, float3 &result) {

   // Recall: a x b = (a2b3 - a3b2)i + (a3b1 - a1b3)j + (a1b2 - a2b1)k
   // Where a = a1 i + a2 j + a3 k and
   //       b = b1 i + b2 j + b3 k

   result.x = d1.y*d2.z - d1.z*d2.y;
   result.y = d1.z*d2.x - d1.x*d2.z;
   result.z = d1.x*d2.y - d1.y*d2.x;

}

__device__
float
cuda_angle_restraint_t::distortion(const float *d_x) const {

   // pre-calculate this?
   float weight =  1.0f/(sigma * sigma);

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;

   float xd_len1 = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd_len1 = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd_len1 = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float xd_len2 = d_x[idx_at_3  ] - d_x[idx_at_2  ];
   float yd_len2 = d_x[idx_at_3+1] - d_x[idx_at_2+1];
   float zd_len2 = d_x[idx_at_3+2] - d_x[idx_at_2+2];

   float len1_sqd = xd_len1*xd_len1 + yd_len1*yd_len1 + zd_len1*zd_len1;
   float len2_sqd = xd_len2*xd_len2 + yd_len2*yd_len2 + zd_len2*zd_len2;

   float len1 = sqrt(len1_sqd);
   float len2 = sqrt(len2_sqd);

   if (len1 < 0.01) { len1 = 0.01; }
   if (len1 < 0.01) { len2 = 0.01; }

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);

   float cos_theta = cuda_lau::dot(d1,d2)/(len1*len2);
   if (cos_theta < -1) cos_theta = -1.0;
   if (cos_theta >  1) cos_theta =  1.0;

   float theta = acos(cos_theta);
   const float rad_to_deg = 180.0f/M_PI;
   const float deg_to_rad = M_PI/180.0f;
   float delta = rad_to_deg * theta - target_value;
   float distortion = weight * delta * delta; 
   // printf("debug angle distortion: theta %f delta %f distortion %f\n", rad_to_deg*theta, delta, distortion);
   return distortion;

}

__device__ __host__
float
cuda_torsion_restraint_t::distortion(const float *d_x) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;
   unsigned int idx_at_4 = 3 * atom_idx_4;

   float xd_len1 = d_x[idx_at_2  ] - d_x[idx_at_1  ];
   float yd_len1 = d_x[idx_at_2+1] - d_x[idx_at_1+1];
   float zd_len1 = d_x[idx_at_2+2] - d_x[idx_at_1+2];

   float xd_len2 = d_x[idx_at_3  ] - d_x[idx_at_2  ];
   float yd_len2 = d_x[idx_at_3+1] - d_x[idx_at_2+1];
   float zd_len2 = d_x[idx_at_3+2] - d_x[idx_at_2+2];

   float xd_len3 = d_x[idx_at_4  ] - d_x[idx_at_3  ];
   float yd_len3 = d_x[idx_at_4+1] - d_x[idx_at_3+1];
   float zd_len3 = d_x[idx_at_4+2] - d_x[idx_at_3+2];

   float len1_sqd = xd_len1*xd_len1 + yd_len1*yd_len1 + zd_len1*zd_len1;
   float len2_sqd = xd_len2*xd_len2 + yd_len2*yd_len2 + zd_len2*zd_len2;
   float len3_sqd = xd_len3*xd_len3 + yd_len3*yd_len3 + zd_len3*zd_len3;

   float len1 = sqrt(len1_sqd);
   float len2 = sqrt(len2_sqd);
   float len3 = sqrt(len3_sqd);

   if (len1 < 0.01) { len1 = 0.01; }
   if (len2 < 0.01) { len2 = 0.01; }
   if (len3 < 0.01) { len3 = 0.01; }

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);
   float3 d3 = make_float3(xd_len3, yd_len3, zd_len3);

   // float E = cuda_lau::dot(d1, cuda_lau::cross(d2, d3));
   float3 cp;
   cuda_lau::cross(d2, d3, cp);
   float E = cuda_lau::dot(d1, cp) * len2;
   float G = -cuda_lau::dot(d1, d3)*len2_sqd + cuda_lau::dot(d1, d2) * cuda_lau::dot(d2, d3);
   float phi = atan2(E, G); // radians

   float phi_0_jk_rad = (CUDART_PI_F/180.0) * phi0_jk;

   float distortion = 0.5 * V_jk * (1.0f - cos(n_jk*(phi - phi_0_jk_rad)));

   if (false)
      printf("torsion distortion %2d %2d %2d %2d V_jk: %f phi0_jk: %8.3f n_jk %f phi %f (deg:) %f phi_0_jk_rad %f -> %f\n",
          atom_idx_1, atom_idx_2, atom_idx_3, atom_idx_4,
          V_jk, phi0_jk, n_jk, phi,
          180.0/3.1415926 * phi,
          phi_0_jk_rad, distortion);

   return distortion;

}


float
cuda_chiral_volume_restraint_t::distortion(const float *d_x) const {

   // a.(bxc)

   float weight =  1.0f/(sigma * sigma);

   unsigned int idx_at_c = 3 * atom_idx_c;
   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;

   float xd_len1 = d_x[idx_at_1  ] - d_x[idx_at_c  ];
   float yd_len1 = d_x[idx_at_1+1] - d_x[idx_at_c+1];
   float zd_len1 = d_x[idx_at_1+2] - d_x[idx_at_c+2];

   float xd_len2 = d_x[idx_at_2  ] - d_x[idx_at_c  ];
   float yd_len2 = d_x[idx_at_2+1] - d_x[idx_at_c+1];
   float zd_len2 = d_x[idx_at_2+2] - d_x[idx_at_c+2];

   float xd_len3 = d_x[idx_at_3  ] - d_x[idx_at_c  ];
   float yd_len3 = d_x[idx_at_3+1] - d_x[idx_at_c+1];
   float zd_len3 = d_x[idx_at_3+2] - d_x[idx_at_c+2];

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);
   float3 d3 = make_float3(xd_len3, yd_len3, zd_len3);

   float3 cp; // fill this
   cuda_lau::cross(d2, d3, cp);
   float cv = cuda_lau::dot(d1, cp);

   float delta = cv - target_chiral_volume;
   float distortion = weight * delta * delta;

   if (false)
      printf("chiral volume distortion %f delta %f\n", cv, delta);

   return distortion;
}


__device__
void
cuda_chiral_volume_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   float weight =  1.0f/(sigma * sigma);

   unsigned int idx_at_c = 3 * atom_idx_c;
   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   unsigned int idx_at_3 = 3 * atom_idx_3;

   float xd_len1 = d_x[idx_at_1  ] - d_x[idx_at_c  ];
   float yd_len1 = d_x[idx_at_1+1] - d_x[idx_at_c+1];
   float zd_len1 = d_x[idx_at_1+2] - d_x[idx_at_c+2];

   float xd_len2 = d_x[idx_at_2  ] - d_x[idx_at_c  ];
   float yd_len2 = d_x[idx_at_2+1] - d_x[idx_at_c+1];
   float zd_len2 = d_x[idx_at_2+2] - d_x[idx_at_c+2];

   float xd_len3 = d_x[idx_at_3  ] - d_x[idx_at_c  ];
   float yd_len3 = d_x[idx_at_3+1] - d_x[idx_at_c+1];
   float zd_len3 = d_x[idx_at_3+2] - d_x[idx_at_c+2];

   float3 d1 = make_float3(xd_len1, yd_len1, zd_len1);
   float3 d2 = make_float3(xd_len2, yd_len2, zd_len2);
   float3 d3 = make_float3(xd_len3, yd_len3, zd_len3);

   float3 cp; // fill this
   cuda_lau::cross(d2, d3, cp);
   float cv = cuda_lau::dot(d1, cp);

   // float delta = cv - target_chiral_volume;
   // float distortion = weight * delta * delta;

   float P0_x_contrib =
      - (d2.y*d3.z - d2.z*d3.y)
      - (d1.z*d3.y - d1.y*d3.z)
      - (d1.y*d2.z - d1.z*d2.y);

   float P0_y_contrib =
      - (d2.z*d3.x - d2.x*d3.z)
      - (d1.x*d3.z - d1.z*d3.x)
      - (d1.z*d2.x - d1.x*d2.z);

   float P0_z_contrib =
      - (d2.x*d3.y - d2.y*d3.x)
      - (d1.y*d3.x - d1.x*d3.y)
      - (d1.x*d2.y - d1.y*d2.x);

   float P1_x_contrib = d2.y*d3.z - d2.z*d3.y;
   float P1_y_contrib = d2.z*d3.x - d2.x*d3.z;
   float P1_z_contrib = d2.x*d3.y - d2.y*d3.x;

   float P2_x_contrib = d1.z*d3.y - d1.y*d3.z;
   float P2_y_contrib = d1.x*d3.z - d1.z*d3.x;
   float P2_z_contrib = d1.y*d3.x - d1.x*d3.y;

   float P3_x_contrib = d1.y*d2.z - d1.z*d2.y;
   float P3_y_contrib = d1.z*d2.x - d1.x*d2.z;
   float P3_z_contrib = d1.x*d2.y - d1.y*d2.x;

   float delta = cv - target_chiral_volume;
   float s = 2.0 * delta * weight;
#if 0
   printf("chiral volume A atom index %d s %f lengths: %f %f %f   %f %f %f    %f %f %f\n", idx_at_c, s, xd_len1, yd_len1, zd_len1, xd_len2, yd_len2, zd_len2, xd_len3, yd_len3, zd_len3);
   printf("chiral volume B atom index %d other atom indices %d %d %d s %f cv %f \n", idx_at_c, idx_at_1, idx_at_2, idx_at_3);
   printf("chiral volume C atom index %d s %f adding these to gradient: x y z %f %f %f \n", atom_idx_c, s, s * P0_x_contrib, s * P0_y_contrib, s * P0_z_contrib);
#endif

   float s_lim = 10000.0;
   if (s >  s_lim) s =  s_lim;
   if (s < -s_lim) s = -s_lim;

   {
      atomicAdd(d_grad+idx_at_c  , s * P0_x_contrib);
      atomicAdd(d_grad+idx_at_c+1, s * P0_y_contrib);
      atomicAdd(d_grad+idx_at_c+2, s * P0_z_contrib);
   }
   {
      atomicAdd(d_grad+idx_at_1  , s * P1_x_contrib);
      atomicAdd(d_grad+idx_at_1+1, s * P1_y_contrib);
      atomicAdd(d_grad+idx_at_1+2, s * P1_z_contrib);
   }
   {
      atomicAdd(d_grad+idx_at_2  , s * P2_x_contrib);
      atomicAdd(d_grad+idx_at_2+1, s * P2_y_contrib);
      atomicAdd(d_grad+idx_at_2+2, s * P2_z_contrib);
   }
   {
      atomicAdd(d_grad+idx_at_3  , s * P3_x_contrib);
      atomicAdd(d_grad+idx_at_3+1, s * P3_y_contrib);
      atomicAdd(d_grad+idx_at_3+2, s * P3_z_contrib);
   }
}

#include "clipper-cuda.cuh"

__device__
cuda_plane_distortion_info_t
cuda_device_plane_restraint_t::cuda_distortion_score_plane_internal(const float *d_x, bool calculate_distortion_flag) const {
   cuda_plane_distortion_info_t info;

   const int &n_atoms = n_atom_index_sigma_pairs;

   if (n_atoms > 0) {
      float sum_x = 0.0;
      float sum_y = 0.0;
      float sum_z = 0.0;
      float dn_atoms = (float) n_atoms;
      for (int i=0; i<n_atoms; i++) {
         int idx = 3 * atom_indices[i];
         sum_x += d_x[idx];
         sum_y += d_x[idx+1];
         sum_z += d_x[idx+2];
      }
      float x_cen = sum_x/dn_atoms;
      float y_cen = sum_y/dn_atoms;
      float z_cen = sum_z/dn_atoms;

      mat33_cuda mat;

      for (int i=0; i<n_atoms; i++) {
	 int idx = 3 * atom_indices[i];
	 float v_idx_0 = d_x[idx  ];
	 float v_idx_1 = d_x[idx+1];
	 float v_idx_2 = d_x[idx+2];
	 mat.m[0][0] += (v_idx_0 - x_cen) * (v_idx_0 - x_cen);
	 mat.m[1][1] += (v_idx_1 - y_cen) * (v_idx_1 - y_cen);
	 mat.m[2][2] += (v_idx_2 - z_cen) * (v_idx_2 - z_cen);
	 mat.m[0][1] += (v_idx_0 - x_cen) * (v_idx_1 - y_cen);
	 mat.m[0][2] += (v_idx_0 - x_cen) * (v_idx_2 - z_cen);
	 mat.m[1][2] += (v_idx_1 - y_cen) * (v_idx_2 - z_cen);
      }
      mat.m[1][0] = mat.m[0][1];
      mat.m[2][0] = mat.m[0][2];
      mat.m[2][1] = mat.m[1][2];

      float3 eigens = mat.eigens(true);

      float abcd[4];
      abcd[0] = mat.m[0][0];
      abcd[1] = mat.m[1][0];
      abcd[2] = mat.m[2][0];

      float sqsum = 1e-20;
      for (int i=0; i<3; i++)
	 sqsum += abcd[i] * abcd[i];
      for (int i=0; i<3; i++)
	 abcd[i] /= sqsum;

      abcd[3] = abcd[0]*x_cen + abcd[1]*y_cen + abcd[2]*z_cen;
      // info.abcd = abcd;
      for(unsigned int i=0; i<4; i++)
	 info.abcd[i] = abcd[i];

      if (calculate_distortion_flag) {
	 float sum_devi = 0.0;
	 for (int i=0; i<n_atoms; i++) {
	    int idx = 3* atom_indices[i];
	    // double val = abcd[0]*gsl_vector_get(v,idx  ) +
	    //              abcd[1]*gsl_vector_get(v,idx+1) +
	    //              abcd[2]*gsl_vector_get(v,idx+2) -
	    //              abcd[3];
	    float val = abcd[0] * d_x[idx] + abcd[1] * d_x[idx+1] + abcd[2] * d_x[idx+2] - abcd[3];
	    float r = val/sigmas[i];
	    sum_devi += r*r;
	 }
	 info.distortion = sum_devi;
      }
   }
   return info;
}

__host__
cuda_plane_distortion_info_t
cuda_host_plane_restraint_t::cuda_distortion_score_plane_internal(const float *d_x, bool calculate_distortion_flag) const {

   cuda_plane_distortion_info_t info;

   int n_atoms = atom_indices_with_sigmas.size();

   if (n_atoms > 0) {
      float sum_x = 0.0;
      float sum_y = 0.0;
      float sum_z = 0.0;
      float dn_atoms = (float) n_atoms;
      for (int i=0; i<n_atoms; i++) {
	 int idx = 3 * atom_indices_with_sigmas[i].first;
	 if (idx < 0) {
	    printf("trapped bad cuda plane restraint atom - should never happen\n");
	 } else {
	    sum_x += d_x[idx];
	    sum_y += d_x[idx+1];
	    sum_z += d_x[idx+2];
	 }
      }
      float x_cen = sum_x/dn_atoms;
      float y_cen = sum_y/dn_atoms;
      float z_cen = sum_z/dn_atoms;

      mat33_cuda mat;

      for (int i=0; i<n_atoms; i++) {
	 int idx = 3 * atom_indices_with_sigmas[i].first;
	 float v_idx_0 = d_x[idx  ];
	 float v_idx_1 = d_x[idx+1];
	 float v_idx_2 = d_x[idx+2];
	 mat.m[0][0] += (v_idx_0 - x_cen) * (v_idx_0 - x_cen);
	 mat.m[1][1] += (v_idx_1 - y_cen) * (v_idx_1 - y_cen);
	 mat.m[2][2] += (v_idx_2 - z_cen) * (v_idx_2 - z_cen);
	 mat.m[0][1] += (v_idx_0 - x_cen) * (v_idx_1 - y_cen);
	 mat.m[0][2] += (v_idx_0 - x_cen) * (v_idx_2 - z_cen);
	 mat.m[1][2] += (v_idx_1 - y_cen) * (v_idx_2 - z_cen);
      }
      mat.m[1][0] = mat.m[0][1];
      mat.m[2][0] = mat.m[0][2];
      mat.m[2][1] = mat.m[1][2];

      float3 eigens = mat.eigens(true);

      float abcd[4];
      abcd[0] = mat.m[0][0];
      abcd[1] = mat.m[1][0];
      abcd[2] = mat.m[2][0];

      float sqsum = 1e-20;
      for (int i=0; i<3; i++)
	 sqsum += abcd[i] * abcd[i];
      for (int i=0; i<3; i++)
	 abcd[i] /= sqsum;

      abcd[3] = abcd[0]*x_cen + abcd[1]*y_cen + abcd[2]*z_cen;
      // info.abcd = abcd;
      for(unsigned int i=0; i<4; i++)
	 info.abcd[i] = abcd[i];

      if (calculate_distortion_flag) {
	 float sum_devi = 0.0;
	 for (int i=0; i<n_atoms; i++) {
	    int idx = 3 * atom_indices_with_sigmas[i].first;
	    // double val = abcd[0]*gsl_vector_get(v,idx  ) +
	    //              abcd[1]*gsl_vector_get(v,idx+1) +
	    //              abcd[2]*gsl_vector_get(v,idx+2) -
	    //              abcd[3];
	    float val = abcd[0] * d_x[idx] + abcd[1] * d_x[idx+1] + abcd[2] * d_x[idx+2] - abcd[3];
	    float r = val/atom_indices_with_sigmas[i].second;
	    sum_devi += r*r;
	 }
	 info.distortion = sum_devi;
      }
   }
   return info;
}

__device__
float
cuda_device_plane_restraint_t::distortion(const float *d_x) const {

   cuda_plane_distortion_info_t info = cuda_distortion_score_plane_internal(d_x, true);
   return info.distortion;
}

__device__
float
cuda_host_plane_restraint_t::distortion(const float *d_x) const {

   cuda_plane_distortion_info_t info = cuda_distortion_score_plane_internal(d_x, true);
   return info.distortion;
}

__device__
void
cuda_device_plane_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   cuda_plane_distortion_info_t info = cuda_distortion_score_plane_internal(d_x, true);
   const int &n_atoms = n_atom_index_sigma_pairs;

   for (int j=0; j<n_atoms; j++) {
      if (true) { // test for fixed atom
	 int idx = 3 * atom_indices[j];
	 float devi_len =
	    info.abcd[0] * d_x[idx  ] +
	    info.abcd[1] * d_x[idx+1] +
	    info.abcd[2] * d_x[idx+2] -
	    info.abcd[3];

	 float weight = 1.0/(sigmas[j] * sigmas[j]);

         atomicAdd(d_grad+idx  , 2.0 * weight * devi_len * info.abcd[0]);
         atomicAdd(d_grad+idx+1, 2.0 * weight * devi_len * info.abcd[1]);
         atomicAdd(d_grad+idx+2, 2.0 * weight * devi_len * info.abcd[2]);
      }
   }
}


__device__
float
cuda_geman_mcclure_restraint_t::distortion(const float *d_x) const {

   // This is now GM only. Write a new function for harmonic restraints

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;
   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   // Let z = (boi - bi)/sigma
   // so S_i = z^2/(1 + alpha * z^2)
   //
   float bl = sqrt(bl_sq);
   float delta = bl - target_value;
   float z = delta/sigma;

   /*
   if (type == GEMAN_MCCLURE) {
      float distortion = z*z/(1+alpha*z*z);
      return distortion;
   } else {
      float distortion = z*z;
      return distortion;
   }
   */

   return z*z/(1+alpha*z*z);

}

__device__
void
cuda_geman_mcclure_restraint_t::add_gradients(const float *d_x, float *d_grad) const {

   unsigned int idx_at_1 = 3 * atom_idx_1;
   unsigned int idx_at_2 = 3 * atom_idx_2;

   float xd = d_x[idx_at_1  ] - d_x[idx_at_2  ];
   float yd = d_x[idx_at_1+1] - d_x[idx_at_2+1];
   float zd = d_x[idx_at_1+2] - d_x[idx_at_2+2];

   float bl_sq = xd*xd + yd*yd + zd*zd;

   float bl = sqrt(bl_sq);
   float delta =  bl - target_value;
   float z = delta/sigma;

   float weight = 1.0f/(sigma * sigma);

   // Let z = (boi - bi)/sigma
   //    S_i = z^2/(1 + alpha * z^2)
   //
   float beta  = 1.0f + alpha * z * z;

   float constant_part_lsq = 2.0*weight * (1.0f - target_value * rsqrt(bl_sq));

   float constant_part = constant_part_lsq / (beta * beta);

   // doing this test is slow (strange/interestingly enough)
   // if (type == HARMONIC)
   // constant_part = constant_part_lsq;

   { // test for fixed flags

      atomicAdd(d_grad+idx_at_1,   constant_part * xd);
      atomicAdd(d_grad+idx_at_1+1, constant_part * yd);
      atomicAdd(d_grad+idx_at_1+2, constant_part * zd);
   }

   {
      atomicAdd(d_grad+idx_at_2,   constant_part * -xd);
      atomicAdd(d_grad+idx_at_2+1, constant_part * -yd);
      atomicAdd(d_grad+idx_at_2+2, constant_part * -zd);
   }
}

__device__
float
cuda_target_pos_restraint_t::distortion(const float *x) const {

   float v = 0.0f;

   int idx = 3 * atom_idx;
   float x_d = atom_pull_target_pos.x - x[idx];
   float y_d = atom_pull_target_pos.y - x[idx+1];
   float z_d = atom_pull_target_pos.z - x[idx+2];
   float d_sqrd = x_d * x_d + y_d * y_d + z_d * z_d;
   const float sigma = 0.03f;
   const float weight = 1.0f/(sigma * sigma);
   return weight * d_sqrd;
}

__device__
void
cuda_target_pos_restraint_t::add_gradients(const float *x, float *d_grad) const {

   int idx = 3 * atom_idx;
   float x_d = atom_pull_target_pos.x - x[idx];
   float y_d = atom_pull_target_pos.y - x[idx+1];
   float z_d = atom_pull_target_pos.z - x[idx+2];
   float d_sqrd = x_d * x_d + y_d * y_d + z_d * z_d;
   const float sigma = 0.03f;
   const float constant_part = 2.0f/(sigma * sigma);

   atomicAdd(d_grad+idx,   constant_part * x_d);
   atomicAdd(d_grad+idx+1, constant_part * y_d);
   atomicAdd(d_grad+idx+2, constant_part * z_d);

}


__device__
void add_electron_density_gradients(int ith, nxmap_cuda *m, const float *d_x, float *d_grad, float map_weight) {

   if (true) {

      int idx = 3 * ith;
      float x = d_x[idx  ];
      float y = d_x[idx+1];
      float z = d_x[idx+2];
      coord_orth_cuda co(x,y,z);
      float atom_weight = 1.0;
      float fudge = 2.0; // the gradients coming out of grad_interp_cubic() are out by a factor of 2 for some reason!

      grad_orth_cuda grad_orth = m->grad_interp_cubic(co);
      float delta[3];

      if (true) {
	 float dvl = m->density_at_point_interp_linear(co);
	 float dvc = m->density_at_point_interp_cubic(co);

	 float micro_step = 0.005;

         for (int i=0; i<3; i++) {
            vec3_cuda xyz(x,y,z);
            vec3_cuda xyz_orig = xyz;
            xyz[i] = xyz_orig[i] - micro_step;
            coord_orth_cuda co1(xyz);
            float v1 = m->density_at_point_interp_cubic(co1);
            xyz[i] = xyz_orig[i] + micro_step;
            coord_orth_cuda co2(xyz);
            float v2 = m->density_at_point_interp_cubic(co2);
            delta[i] = (v2-v1)/(2.0*micro_step);
         }

	 if (false)
	    printf("add_electron_density_gradients() density_grad ith %3d idx %3d xyz %10.5f %10.5f %10.5f dvl: %8.5f dvc: %8.5f grad: is %9.6f %9.6f %9.6f vs %9.6f %9.6f %9.6f \n",
		   ith, idx, x, y, z, dvl, dvc, fudge * grad_orth.dx(), fudge * grad_orth.dy(), fudge * grad_orth.dz(),
		   delta[0], delta[1], delta[2]);

      }

      // atomicAdd(d_grad+idx  , atom_weight * fudge * map_weight * grad_orth.dx());
      // atomicAdd(d_grad+idx+1, atom_weight * fudge * map_weight * grad_orth.dy());
      // atomicAdd(d_grad+idx+2, atom_weight * fudge * map_weight * grad_orth.dz());

      atomicAdd(d_grad+idx  , -atom_weight * map_weight * delta[0]);
      atomicAdd(d_grad+idx+1, -atom_weight * map_weight * delta[1]);
      atomicAdd(d_grad+idx+2, -atom_weight * map_weight * delta[2]);
   }

}




__device__
bool in_range(int test_idx, int idx_start, int idx_end) {
   if (test_idx >= idx_start) {
      if (test_idx < idx_end) {
         return true;
      } else {
         return false;
      }
   } else {
     return false;
   }
}

__device__
int round_32(const int &idx_in) {

   int r = idx_in % 32;
   if (r == 0) {
      return idx_in;
   } else {
      int v = 32 * (idx_in/32 + 1);
      return v;
   }
}


__global__
void test_kernel(unsigned int n_trials, unsigned int n_slots, float *memloc) {

   int idx_thread = threadIdx.x;
   int ith = blockIdx.x * blockDim.x + threadIdx.x;
   printf("test_kernel idx_thread %d ith %d blockIdx.x %d blockDim.x %d\n", idx_thread, ith,
          blockIdx.x, blockDim.x);

   // the launch parameters are blockIdx.x and blockDim.x

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n_trials; ith += blockDim.x * gridDim.x) {
      atomicAdd(memloc, ith);
   }
}



// -----------------------------------------------------------------------------------------
// ---------------------------------- kernel -----------------------------------------------
// -----------------------------------------------------------------------------------------

__global__
void kernel_test_map(const float *d_x, int n_atoms, nxmap_cuda nxm, float *d_y, float *d_distortion_score_electron_density_p, float *d_grad) {

   int idx_thread = threadIdx.x;  // range(blockDim.x)
   int ith = blockIdx.x * blockDim.x + threadIdx.x;

   // in a test kernel, using partial sum only improved over atomicAdd directly to global memory (and then by ~10%) when n was 10,000,000.
   __shared__ double partial_sum;

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n_atoms; ith += blockDim.x * gridDim.x) {
      int idx = 3 * ith;
      float x = d_x[ith  ];
      float y = d_x[ith+1];
      float z = d_x[ith+2];
      coord_orth_cuda co(x,y,z);
      float weight = 1.0;
      float dv = weight * nxm.density_at_point_interp_linear(co);
      // printf("debug in distortion_score_electron_density() ith is %d dv is %f\n", ith, dv);
      atomicAdd(&partial_sum, dv);
      // printf("debug in distortion_score_electron_density() ith is %d sum is %f\n", ith, partial_sum);
   }

   __syncthreads();
   if (idx_thread == 0) {
      // printf("adding partial sum %f to total %f\n", partial_sum, *d_distortion_score_electron_density_p);
      if (partial_sum != 0.0) {  // is this test worth doing?
         atomicAdd(d_distortion_score_electron_density_p, partial_sum);
         printf("after adding partial sum %f total is %f\n", partial_sum, *d_distortion_score_electron_density_p);
      }
   }
   __syncthreads();
   if (ith == 0) {
      printf("distortion score final sum %f\n", *d_distortion_score_electron_density_p);
   }

}


__global__
void kernel_pre_zero_y_and_derivs(float *d_y, float *d_distortion_score_electron_density_p, float *d_grads, int n_atoms) {

   int idx_thread = threadIdx.x;  // range(blockDim.x)
   int ith = blockIdx.x * blockDim.x + threadIdx.x;
   if (ith == 0) {
      *d_y = 0.0;
      *d_distortion_score_electron_density_p = 0.0;
   }

   int n_grads = n_atoms * 3;
   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n_grads; ith += blockDim.x * gridDim.x)
       d_grads[ith] = 0.0;

}

__global__
void kernel_set_lj_epsilon(cuda_restraints_container_t restraints, float e) {

   // Note to self: don't make this, or arguments used like this, const ref.

   int n = restraints.n_nbc_restraints;
   int ith = blockIdx.x * blockDim.x + threadIdx.x;
   if (ith < n) {
      if (false) {
         printf("here in kernel_set_lj_epsilon() with restraints.nbcs 0x%p and n_nbc_restraints %d\n",
             restraints.nbcs, restraints.n_nbc_restraints);
         printf("set epsilon for restraint threadIdx %d blockIdx.x %d blockDim.x %d gridDim.x %d ith %d of %d\n",
                threadIdx.x, blockIdx.x, blockDim.x, gridDim.x, ith, n);
      }
      restraints.nbcs[ith].set_lj_epsilon(e);
   }

}

//
// pass: d_x: pointer to an array of "coordinates" (model),
//
// returns: d_f: target function value for these coordinates
//          d_grad: gradients for these coordinates
__global__
void kernel_combined_f_gradf(int i_round,
                             const float *d_x, int n_atoms,
                             cuda_restraints_container_t restraints,
                             bool do_bonds, bool do_angles, bool do_torsions, bool do_chiral_volumes,
                             bool do_planes, bool do_nbcs, bool do_geman_mcclure_distances,
                             bool do_electron_density,
                             nxmap_cuda nxm, float map_weight,
                             float *d_distortion_score_electron_density_p,
                             float *d_y, float *d_grad) {

   int idx_thread = threadIdx.x;  // range(blockDim.x)
   int ith = blockIdx.x * blockDim.x + threadIdx.x;
   __shared__ float partial_sum;
   partial_sum = 0.0f;

   // no speed up here either ... sigh
   bool shared_atoms = false;
   const int saps = 8*1024;
   __shared__ float shared_atom_positions[saps];  // no cost
   int n_atom_floats = 3 * n_atoms;
   if (n_atom_floats < saps) {
      // printf("using shared atoms\n");
      shared_atoms = true;
      for (int ith = threadIdx.x; ith < n_atom_floats; ith += blockDim.x) {
         shared_atom_positions[ith] = d_x[ith];
      }
      d_x = shared_atom_positions;
   }

   __syncthreads(); // needed? cheap?

   // printf("f_grad_f_kernel idx_thread %d ith %d blockIdx.x %d blockDim.x %d\n", idx_thread, ith, blockIdx.x, blockDim.x);

   // if there are more than one threadblock (gridblock size != 1) the partial_sums don't share
   // the same scope,
   //
   // with a domain, for example, I might want to do that.

   float sum_distortion = 0.0;

   // not clear that round_32() does anything useful
   int idx_bonds_start    = 0;
   int idx_bonds_end      = restraints.n_bond_restraints;
   int idx_angles_start   = round_32(idx_bonds_end);
   int idx_angles_end     = idx_angles_start + restraints.n_angle_restraints;
   int idx_torsions_start = round_32(idx_angles_end);
   int idx_torsions_end   = idx_torsions_start + restraints.n_torsion_restraints;
   int idx_chiral_volumes_start = round_32(idx_torsions_end);
   int idx_chiral_volumes_end = idx_chiral_volumes_start + restraints.n_chiral_volume_restraints;
   int idx_planes_start   = round_32(idx_chiral_volumes_end);
   int idx_planes_end     = idx_planes_start + restraints.n_plane_restraints;
   int idx_gm_start       = round_32(idx_planes_end);
   int idx_gm_end         = idx_gm_start + restraints.n_geman_mcclure_restraints;
   int idx_nbcs_start     = round_32(idx_gm_end);
   int idx_nbcs_end       = idx_nbcs_start + restraints.n_nbc_restraints;
   int idx_restraints_end  = idx_nbcs_end; // just beyond the index of the last restraint, this is for
                                           // an index, not a real count of restraints.
                                           // better rename it to idx_restraints_end

   int idx_density_parts_start = round_32(idx_restraints_end);
   int idx_density_parts_end   = idx_density_parts_start + n_atoms;

   if (do_bonds) {
      const int &n_bond_restraints = restraints.n_bond_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_bonds_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_bonds_start, idx_bonds_end)) {
	    const cuda_bond_restraint_t &rest = restraints.bonds[ith];
	    float distortion = rest.distortion(d_x);
            sum_distortion += distortion;
	 }
      }
   }

   if (do_angles) {
      const int &n_angle_restraints = restraints.n_angle_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_angles_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_angles_start, idx_angles_end)) {
            const cuda_angle_restraint_t &rest = restraints.angles[ith-idx_angles_start];
            float distortion = rest.distortion(d_x);
            sum_distortion += distortion;
         }
      }
   }

   if (do_torsions) {
      const int &n_torsion_restraints = restraints.n_torsion_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_torsions_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_torsions_start, idx_torsions_end)) {
            // printf("Yes to this torsion %d\n", ith);
            const cuda_torsion_restraint_t &rest = restraints.torsions[ith-idx_torsions_start];
            float distortion = rest.distortion(d_x);
            sum_distortion += distortion;
         }
      }
   }

   if (do_chiral_volumes) {
      const int &n_chiral_volume_restraints = restraints.n_chiral_volume_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_chiral_volumes_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_chiral_volumes_start, idx_chiral_volumes_end)) {
            const cuda_chiral_volume_restraint_t &rest = restraints.chiral_volumes[ith-idx_chiral_volumes_start];
            float distortion = rest.distortion(d_x);
            sum_distortion += distortion;
         }
      }
   }

   if (do_planes) {
      const int &n_plane_restraints = restraints.n_plane_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_planes_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_planes_start, idx_planes_end)) {
	    const cuda_device_plane_restraint_t &rest = restraints.planes[ith-idx_planes_start];
	    float distortion = rest.distortion(d_x);
            // printf("plane distortion %d is %f\n", ith, distortion);
            sum_distortion += distortion;
	 }
      }
   }


   if (do_nbcs) {
      const int &n_nbc_restraints = restraints.n_nbc_restraints;
      // unsigned int n_times_for_loop = 0;  22 for a big protein
      // double local_nbc_partial[4];
      // for (unsigned int ii=0; ii<4; ii++) local_nbc_partial[ii] = 0;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_nbcs_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_nbcs_start, idx_nbcs_end)) {
            // n_times_for_loop++;
	    const cuda_nbc_restraint_t &rest = restraints.nbcs[ith-idx_nbcs_start];
	    float distortion = rest.distortion(d_x);
            sum_distortion += distortion;
	 }
      }
   }

   if (do_geman_mcclure_distances) {
      const int &n_geman_mcclure_restraints = restraints.n_geman_mcclure_restraints;
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_gm_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_gm_start, idx_gm_end)) {
	    const cuda_geman_mcclure_restraint_t &rest = restraints.geman_mcclures[ith-idx_gm_start];
	    float distortion = rest.distortion(d_x);
            // printf("geman mcclure ith %d distortion %f\n", ith, distortion);
            sum_distortion += distortion;
         }
      }
   }

   // __syncthreads(); // debugging - no change in speed?

   // ------------------- Gradients -------------------------

   // Note to self, if there are more threads than restraints (very plausible, large number of thread blocks) then
   // it's worth to have ranges for gradients too, idx_bond_gradients_start etc.

   // float local_gradients[4000]; indexed by 3 * n_atoms, i.e. the x,y,z gradients for every atom

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_restraints_end; ith += blockDim.x * gridDim.x) {
      if (do_bonds) {
         if (in_range(ith, idx_bonds_start, idx_bonds_end)) {
            const cuda_bond_restraint_t &rest = restraints.bonds[ith];
            // try returning a list (in this case, just 2 of them) of index,gradient pairs
            rest.add_gradients(d_x, d_grad);
            // rest.get_gradients(d_x, d_grad, local_gradients);
         }
      }
      if (do_angles) {
         if (in_range(ith, idx_angles_start, idx_angles_end)) {
            const cuda_angle_restraint_t &rest = restraints.angles[ith-idx_angles_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
      if (do_torsions) {
         if (in_range(ith, idx_torsions_start, idx_torsions_end)) {
            const cuda_torsion_restraint_t &rest = restraints.torsions[ith-idx_torsions_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
      if (do_chiral_volumes) {
         if (in_range(ith, idx_chiral_volumes_start, idx_chiral_volumes_end)) {
            const cuda_chiral_volume_restraint_t &rest = restraints.chiral_volumes[ith-idx_chiral_volumes_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
      if (do_planes) {
         if (in_range(ith, idx_planes_start, idx_planes_end)) {
            const cuda_device_plane_restraint_t &rest = restraints.planes[ith-idx_planes_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
      if (do_nbcs) {
         if (in_range(ith, idx_nbcs_start, idx_nbcs_end)) {
            const cuda_nbc_restraint_t &rest = restraints.nbcs[ith-idx_nbcs_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
      if (do_geman_mcclure_distances) {
         if (in_range(ith, idx_gm_start, idx_gm_end)) {
            const cuda_geman_mcclure_restraint_t &rest = restraints.geman_mcclures[ith-idx_gm_start];
            rest.add_gradients(d_x, d_grad);
         }
      }
   }

   if (do_electron_density) {
      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_density_parts_end; ith += blockDim.x * gridDim.x) {
         if (in_range(ith, idx_density_parts_start, idx_density_parts_end)) {
            add_electron_density_gradients(ith-idx_density_parts_start, &nxm, d_x, d_grad, map_weight);
         }
      }
   }

   if (do_electron_density) {

      for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < idx_density_parts_end; ith += blockDim.x * gridDim.x) {
	 if (in_range(ith, idx_density_parts_start, idx_density_parts_end)) {
	    int idx = 3 * (ith-idx_density_parts_start);
	    float x = d_x[idx  ];
	    float y = d_x[idx+1];
	    float z = d_x[idx+2];
	    coord_orth_cuda co(x,y,z);
	    float dv = nxm.density_at_point_interp_cubic(co);
	    sum_distortion -= map_weight * dv;
	 }
      }
   }

   // now to add things up.  Using the partial sum doesn't damage (if we sync at the top)
   // but doesn't damage either

   atomicAdd(&partial_sum, sum_distortion);

   __syncthreads(); // no cost

   if (idx_thread == 0)
      atomicAdd(d_y, partial_sum);

   if (false)
      if (ith == 0)
         use_numerical_gradients(restraints, n_atoms, (float *)d_x, d_grad, do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes, do_nbcs, do_geman_mcclure_distances);

#if 0 // debuging gradients.
   if (ith == 0) {
      for (unsigned int i=0; i<n_atoms; i++) {
         printf("gradients %d %f %f %f\n", i, d_grad[3*i], d_grad[3*i+1], d_grad[3*i+2]);
      }
   }
#endif
}

void
gpu_bonds::set_lj_epsilon(const float &e) {

   int n_times = restraints_container.n_nbc_restraints/1024 + 1; // n_nbc / threads/block
   kernel_set_lj_epsilon<<<n_times,1024>>>(restraints_container, e);
   cudaCheckErrors("kernel_set_lj_epsilon set LJ Epsilonfail");
   cudaDeviceSynchronize();
   cudaCheckErrors("cuda kernel fail sync post kernel_set_lj_epsilon()");

}

void
gpu_bonds::f_gradf(const float *x_p, float *y_p, float *grads_p) {

   kernel_pre_zero_y_and_derivs<<<24,512>>>(y_p, d_distortion_score_electron_density_p, grads_p, n_atoms);

   // cudaSetDeviceFlags(cudaDeviceScheduleYield); // does this work?

   // Now I can't launch with 1024 threads. 512 works.
   kernel_combined_f_gradf<<<n_blocks,n_threads_per_block>>>(n_f_gradf, x_p, n_atoms, restraints_container,
                                             do_bonds, do_angles, do_torsions, do_chiral_volumes, do_planes,
                                             do_nbcs, do_geman_mcclure_distances,
                                             do_electron_density,
                                             nxm, map_weight,
                                             d_distortion_score_electron_density_p,
                                             y_p, grads_p);
   cudaCheckErrors("cuda kernel fail kernel_combined_f_gradf()");
   cudaDeviceSynchronize();
   cudaCheckErrors("cuda kernel fail sync post kernel_combined_f_gradf()");
   n_f_gradf++;

}


void
gpu_bonds::print_refinement_flags() const {

   std::cout << "Refinement flags: ";
   if (do_electron_density) std::cout << "density on "; else std::cout << "density off ";
   if (do_bonds) std::cout << "bonds on "; else std::cout << "bonds off ";
   if (do_angles) std::cout << "angles on "; else std::cout << "angles off ";
   if (do_torsions) std::cout << "torsions on "; else std::cout << "torsions off ";
   if (do_planes) std::cout << "planes on "; else std::cout << "planes off ";
   if (do_chiral_volumes) std::cout << "chirals on "; else std::cout << "chirals off ";
   if (do_nbcs) std::cout << "NBCs on "; else std::cout << "NBCs off ";
   if (do_geman_mcclure_distances) std::cout << "GMs on "; else std::cout << "GMs off ";
   std::cout << std::endl;

}

gpu_bonds::gpu_bonds(const cuda_restraints_transfer_container_t &trc, nxmap_cuda d_nxmap, unsigned int n_atoms_in) : cost_function(3*n_atoms_in) {

   do_electron_density = true;
   do_bonds    = true;
   do_angles   = true;
   do_torsions = false;
   do_planes   = true;
   do_nbcs     = true;
   do_chiral_volumes = true;
   do_geman_mcclure_distances = true;
   lj_epsilon = 3.0;

   n_f_gradf = 0; // number of times the kernel was called.

   nxm = d_nxmap;

   n_atoms = n_atoms_in;
   transfer_restraints = trc;
   transfer_restraints_to_device(); // uses transfer_restraints

   print_refinement_flags();

}

#ifndef STANDALONE_INPUTS
cuda_restraints_transfer_container_t convert_to_cuda_restraints(const coot::restraints_container_t &rc) {

   cuda_restraints_transfer_container_t transfer_restraints;

   for (unsigned int i=0; i<rc.size(); i++) {
      const coot::simple_restraint &r = rc[i];
      if (r.restraint_type == coot::BOND_RESTRAINT) {
         cuda_bond_restraint_t cr;
         cr.atom_idx_1 = r.atom_index_1;
         cr.atom_idx_2 = r.atom_index_2;
         cr.target_value = r.target_value;
         cr.sigma     = r.sigma;
         transfer_restraints.bonds.push_back(cr);
         // std::cout << "Bond restraint" << r.atom_index_1 << " " << r.atom_index_2 << " "
         // << r.target_value << std::endl;
      }
      if (r.restraint_type == coot::GEMAN_MCCLURE_DISTANCE_RESTRAINT) {
         cuda_geman_mcclure_restraint_t cr(r.atom_index_1, r.atom_index_2, r.target_value, r.sigma);
         transfer_restraints.geman_mcclures.push_back(cr);
      }
      if (r.restraint_type == coot::NON_BONDED_CONTACT_RESTRAINT) {
         cuda_nbc_restraint_t cr(r.atom_index_1, r.atom_index_2, r.target_value);
         // cr.set_lj_epsilon(0.20);
         transfer_restraints.nbcs.push_back(cr);
      }
      if (r.restraint_type == coot::ANGLE_RESTRAINT) {
          cuda_angle_restraint_t cr(r.atom_index_1, r.atom_index_2, r.atom_index_3, r.target_value, r.sigma);
          transfer_restraints.angles.push_back(cr);
      }
      if (r.restraint_type == coot::TORSION_RESTRAINT) {
          cuda_torsion_restraint_t cr(r.atom_index_1, r.atom_index_2, r.atom_index_3, r.atom_index_4, r.target_value, r.sigma, r.periodicity);
          if (true) { // testing torsion
             clipper::Coord_orth p1 = coot::co(rc.get_atom(r.atom_index_1));
             clipper::Coord_orth p2 = coot::co(rc.get_atom(r.atom_index_2));
             clipper::Coord_orth p3 = coot::co(rc.get_atom(r.atom_index_3));
             clipper::Coord_orth p4 = coot::co(rc.get_atom(r.atom_index_4));
             double t = clipper::Util::rad2d(clipper::Coord_orth::torsion(p1, p2, p3, p4));
             if (false)
                std::cout << "new cuda_torsion_restraint_t: "
                          << r.atom_index_1 << " "
                          << r.atom_index_2 << " "
                          << r.atom_index_3 << " "
                          << r.atom_index_4 << " "
                          << " V_jk: " << cr.V_jk << " n_jk: " << cr.n_jk << " phi0_jk: " << cr.phi0_jk
                          << " clipper_torsion " << t << std::endl;
             transfer_restraints.torsions.push_back(cr);
         }
      }
      if (r.restraint_type == coot::CHIRAL_VOLUME_RESTRAINT) {
         cuda_chiral_volume_restraint_t cr(r.atom_index_centre, r.atom_index_1, r.atom_index_2, r.atom_index_3, r.target_chiral_volume, r.sigma);
         transfer_restraints.chiral_volumes.push_back(cr);
      }
      if (r.restraint_type == coot::PLANE_RESTRAINT) {
         cuda_host_plane_restraint_t cr;
         for (unsigned int iat=0; iat<r.plane_atom_index.size(); iat++) {
            std::pair<unsigned int, float> p(r.plane_atom_index[iat].first, r.plane_atom_index[iat].second);
            cr.atom_indices_with_sigmas.push_back(p);
         }
         transfer_restraints.planes.push_back(cr);
      }
   }
   return transfer_restraints;
}
#endif // STANDALONE_INPUTS

void
gpu_bonds::transfer_restraints_to_device() {

   const cuda_restraints_transfer_container_t &trc = transfer_restraints;
   restraints_container.n_bond_restraints    = trc.bonds.size();
   restraints_container.n_angle_restraints   = trc.angles.size();
   restraints_container.n_torsion_restraints = trc.torsions.size();
   restraints_container.n_chiral_volume_restraints = trc.chiral_volumes.size();
   restraints_container.n_nbc_restraints     = trc.nbcs.size();
   restraints_container.n_plane_restraints   = trc.planes.size();
   restraints_container.n_geman_mcclure_restraints = trc.geman_mcclures.size();

   // -------------- bonds ---------------------
   int n_bonds = trc.bonds.size();

   // std::cout << "debug:: a cuda_bond_restraint_t has size " << sizeof(cuda_bond_restraint_t) << std::endl;
   // std::cout << "debug:: allocating " << n_bonds * sizeof(cuda_bond_restraint_t) << " bytes for bond restraints" << std::endl;

   cuda_bond_restraint_t *r_p = new cuda_bond_restraint_t[n_bonds];
   if (!r_p) return;
   // now fill r_p
   for (unsigned int i=0; i<n_bonds; i++)
      r_p[i] = trc.bonds[i];

   cudaMalloc(&(restraints_container.bonds), n_bonds * sizeof(cuda_bond_restraint_t));
   cudaCheckErrors("cudaMalloc bonds fail");
   cudaMemcpy(restraints_container.bonds, r_p, n_bonds * sizeof(cuda_bond_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy bonds fail");
   delete [] r_p;
   r_p = 0;

   if (false) {
      cuda_bond_restraint_t *receive_p = new cuda_bond_restraint_t[n_bonds];
      cudaMemcpy(receive_p, restraints_container.bonds, n_bonds * sizeof(cuda_bond_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_bonds; i++) {
          const cuda_bond_restraint_t &br = receive_p[i];
          std::cout << " received bond " << i << " : " << br.atom_idx_1 << " " << br.atom_idx_2 << " "
                    << br.target_value << " " << br.sigma << std::endl;
      }
      delete [] receive_p;
   }

   // -------------- nbc ---------------------
   int n_nbcs = restraints_container.n_nbc_restraints;
   cuda_nbc_restraint_t *r_nbc_p = new cuda_nbc_restraint_t[n_nbcs];
   if (!r_nbc_p) return;
   for (unsigned int i=0; i<n_nbcs; i++)
      r_nbc_p[i] = trc.nbcs[i];
   cudaMalloc(&(restraints_container.nbcs), n_nbcs * sizeof(cuda_nbc_restraint_t));
   cudaCheckErrors("cudaMalloc nbc fail");
   cudaMemcpy(restraints_container.nbcs, r_nbc_p, n_nbcs * sizeof(cuda_nbc_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy nbc fail");
   delete [] r_nbc_p;
   r_nbc_p = 0;
   if (false) {
      cuda_nbc_restraint_t *receive_p = new cuda_nbc_restraint_t[n_nbcs];
      cudaMemcpy(receive_p, restraints_container.nbcs, n_nbcs * sizeof(cuda_nbc_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_nbcs; i++) {
          const cuda_nbc_restraint_t &br = receive_p[i];
          std::cout << " received nbc " << i << " : " << br.atom_idx_1 << " " << br.atom_idx_2 << " " << br.target_value << std::endl;
      }
      delete [] receive_p;
   }

   // -------------- geman mcclures ---------------------
   int n_geman_mcclures = restraints_container.n_geman_mcclure_restraints;
   unsigned int gm_restraints_malloc_size = n_geman_mcclures * sizeof(cuda_geman_mcclure_restraint_t);
   std::cout << "Transfering " << n_geman_mcclures << " Geman McClure Restraints," << " size " << gm_restraints_malloc_size << " bytes " << std::endl;
   cuda_geman_mcclure_restraint_t *r_gm_p = new cuda_geman_mcclure_restraint_t[n_geman_mcclures];
   if (!r_gm_p) return;
   for (unsigned int i=0; i<n_geman_mcclures; i++)
      r_gm_p[i] = trc.geman_mcclures[i];
   cudaMalloc(&(restraints_container.geman_mcclures), gm_restraints_malloc_size);
   cudaCheckErrors("cudaMalloc GM fail");
   cudaMemcpy(restraints_container.geman_mcclures, r_gm_p, n_geman_mcclures * sizeof(cuda_geman_mcclure_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy GM fail");
   delete [] r_gm_p;
   r_gm_p = 0;
   if (false) {
      cuda_geman_mcclure_restraint_t *receive_p = new cuda_geman_mcclure_restraint_t[n_geman_mcclures];
      cudaMemcpy(receive_p, restraints_container.geman_mcclures, n_geman_mcclures * sizeof(cuda_geman_mcclure_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_geman_mcclures; i++) {
          const cuda_geman_mcclure_restraint_t &br = receive_p[i];
          std::cout << " received GM " << i << " : " << br.atom_idx_1 << " " << br.atom_idx_2 << " " << br.target_value << " sigma "
                     << br.sigma << " alpha " << br.alpha << std::endl;
      }
      delete [] receive_p;
   }

   // -------------- angles ---------------------
   int n_angles = trc.angles.size();

   // std::cout << "debug:: a cuda_angle_restraint_t has size " << sizeof(cuda_angle_restraint_t) << std::endl;
   // std::cout << "debug:: allocating " << n_angles * sizeof(cuda_angle_restraint_t) << " bytes for angle restraints"
   //     << std::endl;

   cuda_angle_restraint_t *r_angle_p = new cuda_angle_restraint_t[n_angles];
   if (!r_angle_p) return;
   // now fill r_p
   for (unsigned int i=0; i<n_angles; i++)
      r_angle_p[i] = trc.angles[i];

   cudaMalloc(&(restraints_container.angles), n_angles * sizeof(cuda_angle_restraint_t));
   cudaCheckErrors("cudaMalloc angles fail");
   cudaMemcpy(restraints_container.angles, r_angle_p, n_angles * sizeof(cuda_angle_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy angles fail");
   delete [] r_angle_p;
   r_angle_p = 0;

   if (false) {
      cuda_angle_restraint_t *receive_p = new cuda_angle_restraint_t[n_angles];
      cudaMemcpy(receive_p, restraints_container.angles, n_angles * sizeof(cuda_angle_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_angles; i++) {
          const cuda_angle_restraint_t &ar = receive_p[i];
          std::cout << " received angle " << i << " : " << ar.atom_idx_1 << " " << ar.atom_idx_2 << " " << ar.atom_idx_3
                    << " "<< ar.target_value << " " << ar.sigma << std::endl;
      }
      delete [] receive_p;
   }

   // -------------- torsions -----------------------
   int n_torsions = trc.torsions.size();
   cuda_torsion_restraint_t *r_torsion_p = new cuda_torsion_restraint_t[n_torsions];
   if (! r_torsion_p) return;
   for (unsigned int i=0; i< n_torsions; i++)
      r_torsion_p[i] = trc.torsions[i];
   cudaMalloc(&(restraints_container.torsions), n_torsions * sizeof(cuda_torsion_restraint_t));
   cudaCheckErrors("cudaMalloc torsions fail");
   cudaMemcpy(restraints_container.torsions, r_torsion_p, n_torsions * sizeof(cuda_torsion_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy torsions fail");
   delete [] r_angle_p;
   r_angle_p = 0;
   
   if (false) {
      cuda_torsion_restraint_t *receive_p = new cuda_torsion_restraint_t[n_torsions];
      cudaMemcpy(receive_p, restraints_container.torsions, n_torsions * sizeof(cuda_torsion_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_torsions; i++) {
          const cuda_torsion_restraint_t &ar = receive_p[i];
          std::cout << " received torsion " << i << " : " << ar.atom_idx_1 << " " << ar.atom_idx_2 << " " << ar.atom_idx_3 << " " << ar.atom_idx_4
                    << " " << ar.V_jk << " " << ar.n_jk << " " << ar.phi0_jk << std::endl;
      }
      delete [] receive_p;
   }

   // -------------- chiral_volumes ---------------------
   int n_chiral_volumes = trc.chiral_volumes.size();

   // std::cout << "debug:: a cuda_chiral_volume_restraint_t has size " << sizeof(cuda_chiral_volume_restraint_t) << std::endl;
   // std::cout << "debug:: allocating " << n_chiral_volumes * sizeof(cuda_chiral_volume_restraint_t) << " bytes for chiral_volume restraints"
   //    << std::endl;

   cuda_chiral_volume_restraint_t *r_chiral_volume_p = new cuda_chiral_volume_restraint_t[n_chiral_volumes];
   if (!r_chiral_volume_p) return;
   // now fill r_p
   for (unsigned int i=0; i<n_chiral_volumes; i++)
      r_chiral_volume_p[i] = trc.chiral_volumes[i];

   cudaMalloc(&(restraints_container.chiral_volumes), n_chiral_volumes * sizeof(cuda_chiral_volume_restraint_t));
   cudaCheckErrors("cudaMalloc chiral_volumes fail");
   cudaMemcpy(restraints_container.chiral_volumes, r_chiral_volume_p, n_chiral_volumes * sizeof(cuda_chiral_volume_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy chiral_volumes fail");
   delete [] r_chiral_volume_p;
   r_chiral_volume_p = 0;

   if (false) {
      cuda_chiral_volume_restraint_t *receive_p = new cuda_chiral_volume_restraint_t[n_chiral_volumes];
      cudaMemcpy(receive_p, restraints_container.chiral_volumes, n_chiral_volumes * sizeof(cuda_chiral_volume_restraint_t), cudaMemcpyDeviceToHost);
      for (unsigned int i=0; i<n_chiral_volumes; i++) {
          const cuda_chiral_volume_restraint_t &ar = receive_p[i];
          std::cout << " received chiral_volume " << i << " : " << ar.atom_idx_c << " " << ar.atom_idx_1 << " " << ar.atom_idx_2 << " " << ar.atom_idx_3
		    << " "<< ar.target_chiral_volume << " " << ar.sigma << std::endl;
      }
      delete [] receive_p;
   }


   // -------------- planes ---------------------
   int n_planes = trc.planes.size();

   cuda_device_plane_restraint_t *r_device_plane_p = new cuda_device_plane_restraint_t[n_planes];
   if (!r_device_plane_p) return;
   // now fill r_p, converting from host to device
   for (unsigned int i=0; i<n_planes; i++) {
      r_device_plane_p[i] = trc.planes[i];
   }

   cudaMalloc(&(restraints_container.planes), n_planes * sizeof(cuda_device_plane_restraint_t));
   cudaCheckErrors("cudaMalloc planes fail");
   cudaMemcpy(restraints_container.planes, r_device_plane_p, n_planes * sizeof(cuda_device_plane_restraint_t), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMmemcpy planes fail");
   delete [] r_device_plane_p;
   r_device_plane_p = 0;

   if (false) {
      cuda_device_plane_restraint_t *receive_p = new cuda_device_plane_restraint_t[n_planes];
      cudaMemcpy(receive_p, restraints_container.planes, n_planes * sizeof(cuda_device_plane_restraint_t), cudaMemcpyDeviceToHost);
      std::cout << "n_planes = " << n_planes << std::endl;
      for (unsigned int i=0; i<n_planes; i++) {
	 const cuda_device_plane_restraint_t &ar = receive_p[i];
	 // sigma and atom_indices are device pointers. They need to be converted.
	 std::cout << " received plane " << i << " : n-atoms-in-plane: " << ar.n_atom_index_sigma_pairs << std::endl;
	 float s;
	 int idx;
	 for (unsigned int iat=0; iat<ar.n_atom_index_sigma_pairs; iat++) {
	    // std::cout << "   " << ar.atom_indices[iat] << " " << ar.sigmas[iat] << std::endl;
	    cudaMemcpy(&idx, &(ar.atom_indices[iat]), sizeof(int), cudaMemcpyDeviceToHost);
	    cudaCheckErrors("cudaMmemcpy planes atom iat idx fail");
	    cudaMemcpy(&s, &(ar.sigmas[iat]), sizeof(float), cudaMemcpyDeviceToHost);
	    cudaCheckErrors("cudaMmemcpy planes atom iat sigma fail");
	    std::cout << "  idx " << idx << " sigma " << s << "\n";
	 }
	 std::cout << std::endl;
      }
      delete [] receive_p;
   }

}

#ifndef STANDALONE_INPUTS
void
write_atoms(const coot::restraints_container_t &rc, const std::string &file_name) {

   std::ofstream f(file_name);
   if (f) {
      unsigned int n = rc.get_n_atoms();
      for (std::size_t i=0; i<n; i++) {
	 mmdb::Atom *at = rc.get_atom(i);
	 f << at->x << "  " << at->y << "  " << at->z << "\n";
      }
      f.close();
   }
}
#endif // STANDALONE_INPUTS


gpu_bonds::~gpu_bonds() {

  // clear up d_restraints_p and device coords

  // clear up device restraints

}

__global__
void
kernel_get_density_score_sum_at_atoms(const float *d_x, int n_atoms, nxmap_cuda nxm, float *d_distortion_score_electron_density_p) {

   *d_distortion_score_electron_density_p = 0.0;

   __shared__ double partial_sum;
   partial_sum = 0.0f;
   int idx_thread = threadIdx.x;

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n_atoms; ith += blockDim.x * gridDim.x) {
      int idx = 3 * ith;
      float x = d_x[idx  ];
      float y = d_x[idx+1];
      float z = d_x[idx+2];
      coord_orth_cuda co(x,y,z);
      float weight = 1.0f;
      float dv = weight * nxm.density_at_point_interp_cubic(co);
      atomicAdd(&partial_sum, dv);
   }

   __syncthreads(); // safe, but needed? (I think it is, otherwise idx_thread==0 will try to access the
                    // memory in partial sum before all the other threads have added their bit).
                    // Back again, Yes, needed.
   if (idx_thread == 0) {
      if (partial_sum != 0.0f) {  // is this test worth doing?
         atomicAdd(d_distortion_score_electron_density_p, partial_sum);
      }
   }
}

float
get_density_score_sum_at_atoms(const float *d_x, int n_atoms, nxmap_cuda nxm) {

   float ed_host = 0.0;
   float *d_distortion_score_electron_density_p;
   typedef clipper::NXmap<float>::Map_reference_index NRI;
   cudaMalloc(&d_distortion_score_electron_density_p, sizeof(float));
   cudaCheckErrors("cudaMalloc ed distortion fail");
   cudaMemcpy(d_distortion_score_electron_density_p, &ed_host, sizeof(float), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMemcpy ed distortion fail");
   kernel_get_density_score_sum_at_atoms<<<1,32>>>(d_x, n_atoms, nxm, d_distortion_score_electron_density_p);
   cudaCheckErrors("kernel get_density_score_sum_at_atoms fail");
   cudaMemcpy(&ed_host, d_distortion_score_electron_density_p, sizeof(float), cudaMemcpyDeviceToHost);
   cudaCheckErrors("cudaMemcpy get_density_score_sum_at_atoms fail");
   std::cout << "debug:: in get_density_score_sum_at_atoms() density score " << ed_host << " which is "
             << ed_host/static_cast<float>(n_atoms) << " per atom " << n_atoms << std::endl;
   return ed_host;
}




#ifdef STANDALONE_INPUTS
void cuda_minimize(const cuda_restraints_transfer_container_t &transfer_restraints,
                   const atoms_container_t &ac,
                   nxmap_cuda h_map,
                   float map_weight,
                   unsigned int n_blocks,
                   unsigned int n_threads_per_block) {

   {
      nxmap_cuda d_nx(h_map, true); // transfer pointed to data from host to device
      atoms_container_t ac = cuda_minimize_inner(transfer_restraints, ac, d_nx, map_weight, n_blocks, n_threads_per_block, false);
   }
}
#endif // STANDALONE_INPUTS


#ifndef STANDALONE_INPUTS
void cuda_minimize(const coot::restraints_container_t &restraints,
                   atom_selection_container_t asc,
                   const clipper::Xmap<float> &xmap,
                   float map_weight,
                   unsigned int nb, unsigned int nt,
                   bool write_data_for_standalone) {

   std::string pdb_output_file_name("minimized.pdb"); // pass this
   std::string cif_output_file_name("minimized.cif"); // pass this

   int n_atoms = restraints.get_n_atoms();
   std::cout <<  "INFO:: cuda_minimize() with n_atoms "    << n_atoms    << std::endl;
   std::cout <<  "INFO:: cuda_minimize() with map_weight " << map_weight << std::endl;
   std::cout <<  "INFO:: cuda_minimize() with n_blocks  "  << nb         << std::endl;
   std::cout <<  "INFO:: cuda_minimize() with n_threads "  << nt         << std::endl;

   bool debug_coords = false;

   // capture the throw std::length_error from make_nxmap
   try {

      float border_width = 3.0;

      std::pair<clipper::Coord_orth, clipper::Coord_orth> mol_extents = coot::util::extents(asc.mol, asc.SelectionHandle);
      std::pair<clipper::Coord_frac, clipper::Coord_frac> fc_pair =
         coot::util::find_struct_fragment_coord_fracs_v2(mol_extents, xmap.cell());
      double occupied_fraction = (fc_pair.second.u() - fc_pair.first.u()) *
                                 (fc_pair.second.v() - fc_pair.first.v()) *
                                 (fc_pair.second.w() - fc_pair.first.w());
      std::cout << "INFO:: mol orthogonal extents " << mol_extents.first.format() << " to " << mol_extents.second.format() << "\n"
                << "       fractional extents " << fc_pair.first.format() << " to " << fc_pair.second.format() << "\n"
                << "       occupied: " << std::setprecision(2) << std::fixed << (100 * occupied_fraction) << "% "
                << " empty: " << std::setprecision(2) << std::fixed << 100 * (1.0 - occupied_fraction) << "%\n"
                << "       making map with border width " << border_width << std::endl;

      // This used to be slow, now it is threaded :-)
      clipper::NXmap<float> nxmap = coot::util::make_nxmap(xmap, asc, border_width);

      std::cout << "INFO:: grid: " << nxmap.grid().format() << std::endl;

      std::cout << "INFO:: Making nxmap_cuda on device" << std::endl;
      nxmap_cuda nxc(nxmap, true); // transform and put it to the device

      if (false) {
         std::string map_out_file_name("xmap-out.map");
         try {
            clipper::CCP4MAPfile mapout;
            mapout.open_write(map_out_file_name);
            mapout.set_cell(xmap.cell());
            mapout.export_nxmap(nxmap);
            mapout.close_write();
         }
         catch (const clipper::Message_base &except) {
            std::cout << "WARNING:: Failed to write " << map_out_file_name  << std::endl;
         }
      }

      if (write_data_for_standalone) {
         nxmap_cuda h_map(nxmap, false); // construct in host memory
         h_map.write_map("standalone-map.list");

         // delete h_map
      }

      cuda_restraints_transfer_container_t transfer_restraints = convert_to_cuda_restraints(restraints);

      atoms_container_t ac(asc.mol);
      if (false) {
         for(unsigned int i=0; i<ac.size(); i++) {
            std::cout << "begin pos " << i << " " << ac[i].xyz[0] << " " << ac[i].xyz[1] << " " << ac[i].xyz[2] << "\n";
         }
      }

      atoms_container_t ac_new = cuda_minimize_inner(transfer_restraints, ac, nxc, map_weight, nb, nt,
                                                     write_data_for_standalone);
      ac_new.transfer_to(asc);
      // do this in the calling function now
      // asc.mol->WritePDBASCII(pdb_output_file_name.c_str());
      // asc.mol->WriteCIFASCII(cif_output_file_name.c_str());
   }

   catch (const std::length_error &le) {
      std::cout << "Error: Can't make an NXMap that big: length_error " << le.what() << std::endl;
   }
}
#endif // STANDALONE_INPUTS



void monitor_moving_atoms(float *x, float *d_x, const atoms_container_t &atom_container) {

   unsigned int n_atoms = atom_container.size();
   float dist_crit = 1.0;
   float dist_crit_sq = dist_crit * dist_crit;

   // atom_container is where there were. x is some memory on the host, d_x is where they
   // are on the device. Let's copy them from d_x to x and compair them to the positions
   // in atom_container.

   // copy the atom positions from the device back into x.
   cudaMemcpy(x, d_x, n_atoms * 3  * sizeof(float), cudaMemcpyDeviceToHost);
   if (false)
      for (unsigned int i=0; i<atom_container.size(); i++)
         std::cout << "Resulting coords: " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;

   // now fill the return atoms
   atoms_container_t atoms_current = atom_container;
   for (unsigned int i=0; i<atom_container.size(); i++) {
      atoms_current.positions[i].xyz[0] = x[3*i];
      atoms_current.positions[i].xyz[1] = x[3*i+1];
      atoms_current.positions[i].xyz[2] = x[3*i+2];
   }

   for (unsigned int i=0; i<atom_container.size(); i++) {
      float dd = atoms_container_t::pos_t::distsq(atom_container[i], atoms_current[i]);
      if (dd > dist_crit_sq) {
         std::cout << "Weee! " << i << " " << sqrt(dd) << " " << atom_container.spec(i) << std::endl;
      }
   }

}

atoms_container_t
cuda_minimize_inner(const cuda_restraints_transfer_container_t &transfer_restraints,
                    const atoms_container_t &atom_container,
                    const nxmap_cuda &d_nxc,
                    float map_weight,
                    unsigned int nb, unsigned int nt,
                    bool write_data_for_standalone) {

   atoms_container_t atoms_for_return;
   {
      gpu_bonds gb(transfer_restraints, d_nxc, atom_container.size());

#ifndef STANDALONE_INPUTS
      if (write_data_for_standalone) {
         transfer_restraints.write("standalone-restraints.list");
         atom_container.write_atoms("standalone-atoms.list");
      }
#endif // STANDALONE_INPUTS

      // electron density distortion score holder - needs to be cleared on destruction FIXME
      float ed_host = 0.0;
      cudaMalloc(&gb.d_distortion_score_electron_density_p, sizeof(float));
      cudaCheckErrors("cudaMalloc ed distortion fail");
      cudaMemcpy(gb.d_distortion_score_electron_density_p, &ed_host, sizeof(float), cudaMemcpyHostToDevice);
      cudaCheckErrors("cudaMemcpy ed distortion fail");

      // allocate host memory for atoms:
      unsigned int n_atoms = atom_container.size();
      std::cout << "allocating host memory for " << atom_container.size() << " atoms" << std::endl;
      float *x = new float[atom_container.size() * 3];
      if (!x) return atoms_for_return;
      // allocate d_x on the device
      float *d_x;
      cudaMalloc(&d_x, n_atoms * 3 * sizeof(float));
      // now fill x
      for (unsigned int i=0; i<atom_container.size(); i++) {
         x[3*i  ] = atom_container[i].xyz[0];
         x[3*i+1] = atom_container[i].xyz[1];
         x[3*i+2] = atom_container[i].xyz[2];
      }
      if (false)
         for (unsigned int i=0; i<atom_container.size(); i++)
            std::cout << "Input coords: " << i << " " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;

      // copy x to d_x
      cudaMemcpy(d_x, x, n_atoms * 3 * sizeof(float), cudaMemcpyHostToDevice);

      float ed_start = get_density_score_sum_at_atoms(d_x, n_atoms, d_nxc);

      gb.set_lj_epsilon(0.002);
      gb.set_map_weight(map_weight);
      gb.set_launch_parameter(nb, nt);
      lbfgs minimizer_1(gb);
      minimizer_1.setGradientEpsilon(0.03);

      float sum_bond_distortion    = gb.transfer_restraints.sum_bond_distortion(x); // host memory
      float sum_angle_distortion   = gb.transfer_restraints.sum_angle_distortion(x);
      float sum_torsion_distortion = gb.transfer_restraints.sum_torsion_distortion(x);
      float sum_chiral_volume_distortion = gb.transfer_restraints.sum_chiral_volume_distortion(x);
      float sum_nbc_distortion           = gb.transfer_restraints.sum_nbc_distortion(x);
      float sum_plane_distortion         = gb.transfer_restraints.sum_plane_distortion(x);
      float sum_geman_mcclure_distortion = gb.transfer_restraints.sum_geman_mcclure_distortion(x);

      std::cout << "pre-minimize bond   distortion " << sum_bond_distortion  << "   " << sqrt(sum_bond_distortion/static_cast<float>(gb.transfer_restraints.bonds.size()))  << std::endl;
      std::cout << "pre-minimize angle  distortion " << sum_angle_distortion << "   " << sqrt(sum_angle_distortion/static_cast<float>(gb.transfer_restraints.angles.size())) << std::endl;
      if (gb.transfer_restraints.torsions.size() > 0)
         std::cout << "pre-minimize torsion distortion " << sum_torsion_distortion << "   " << sqrt(sum_torsion_distortion/static_cast<float>(gb.transfer_restraints.torsions.size())) << std::endl;
      std::cout << "pre-minimize chiral distortion " << sum_chiral_volume_distortion << "   "
                << sqrt(sum_chiral_volume_distortion/static_cast<float>(gb.transfer_restraints.chiral_volumes.size())) << std::endl;
      std::cout << "pre-minimize plane  distortion " << sum_plane_distortion << "   " << sqrt(sum_plane_distortion/static_cast<float>(gb.transfer_restraints.planes.size())) << std::endl;
      std::cout << "pre-minimize nbc    distortion " << sum_nbc_distortion   << "   " << sum_nbc_distortion/static_cast<float>(gb.transfer_restraints.nbcs.size()) << std::endl;
      std::cout << "pre-minimize GM     distortion " << sum_geman_mcclure_distortion << "   "
                << sum_geman_mcclure_distortion/static_cast<float>(gb.transfer_restraints.n_geman_mcclures()) << " n = " << gb.transfer_restraints.n_geman_mcclures() << std::endl;
      if (gb.transfer_restraints.torsions.size() > 0) { 
      } else {
        std::cout << "No torsions\n";
      }

      if (true) {
         std::cout << "Minimize..." << std::endl;
         minimizer_1.setMaxIterations(100);
         // At some stage, I will want to minimize for a "few" cycles, not all the way, use max_iter
         auto tp_0 = std::chrono::high_resolution_clock::now();
         lbfgs::status s1 = minimizer_1.minimize(d_x);
         auto tp_1 = std::chrono::high_resolution_clock::now();

         monitor_moving_atoms(x, d_x, atom_container);

         gb.set_lj_epsilon(3.0);
         lbfgs minimizer_2(gb);
         minimizer_2.setMaxIterations(100000);
         lbfgs::status s2 = minimizer_2.minimize(d_x);
         auto tp_2 = std::chrono::high_resolution_clock::now();
         auto d10 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_1 - tp_0).count();
         auto d21 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_2 - tp_1).count();
         std::cout << "info:: " << d10 << " and " << d21 << " milliseconds for " << gb.get_n_f_gradf() << " kernel calls\n";

         std::cout << "lbfgs status " << s1 << " " << s2 << std::endl;
         switch (s2) {
           case lbfgs::LBFGS_BELOW_GRADIENT_EPS:
                std::cout << "LBFGS_BELOW_GRADIENT_EPS" << std::endl;
                break;
           case lbfgs::LBFGS_REACHED_MAX_ITER:
                std::cout << "LBFGS_REACHED_MAX_ITER" << std::endl;
                break;
           case lbfgs::LBFGS_REACHED_MAX_EVALS:
                std::cout << "LBFGS_REACHED_MAX_EVALS" << std::endl;
                break;
           case lbfgs::LBFGS_LINE_SEARCH_FAILED:
                std::cout << "LBFGS_LINE_SEARCH_FAILED" << std::endl;
                break;
         }
      }

      // copy the atom positions from the device back into x.
      cudaMemcpy(x, d_x, n_atoms * 3  * sizeof(float), cudaMemcpyDeviceToHost);
      if (false)
         for (unsigned int i=0; i<atom_container.size(); i++)
            std::cout << "Resulting coords: " << x[3*i] << " " << x[3*i+1] << " " << x[3*i+2] << std::endl;

      // now fill the return atoms
      atoms_for_return = atom_container;
      for (unsigned int i=0; i<atom_container.size(); i++) {
         atoms_for_return.positions[i].xyz[0] = x[3*i];
         atoms_for_return.positions[i].xyz[1] = x[3*i+1];
         atoms_for_return.positions[i].xyz[2] = x[3*i+2];
      }

      // transfer back the electron density score, currently not filled.
      cudaMemcpy(&ed_host, gb.d_distortion_score_electron_density_p, sizeof(float), cudaMemcpyDeviceToHost);
      cudaCheckErrors("cudaMemcpy ed distortion fail");

      float ed_end = get_density_score_sum_at_atoms(d_x, n_atoms, d_nxc);
      std::cout << "Post-fit electron density score " << ed_end << " which is " << ed_end/static_cast<float>(n_atoms)
                << " per atom" << std::endl;

      sum_bond_distortion  = gb.transfer_restraints.sum_bond_distortion(x); // host memory
      sum_angle_distortion = gb.transfer_restraints.sum_angle_distortion(x);
      sum_torsion_distortion = gb.transfer_restraints.sum_torsion_distortion(x);
      sum_plane_distortion = gb.transfer_restraints.sum_plane_distortion(x);
      sum_nbc_distortion   = gb.transfer_restraints.sum_nbc_distortion(x);
      sum_chiral_volume_distortion = gb.transfer_restraints.sum_chiral_volume_distortion(x);
      sum_geman_mcclure_distortion = gb.transfer_restraints.sum_geman_mcclure_distortion(x);

      std::string refined_bonds;
      std::string refined_angles;
      std::string refined_torsions;
      std::string refined_planes;
      std::string refined_nbcs;
      std::string refined_chiral_volumes;
      std::string refined_GMs;

      if (gb.do_bonds)    refined_bonds    = " refined";
      if (gb.do_angles)   refined_angles   = " refined";
      if (gb.do_torsions) refined_torsions = " refined";
      if (gb.do_planes)   refined_planes   = " refined";
      if (gb.do_nbcs)     refined_nbcs     = " refined";
      if (gb.do_chiral_volumes) refined_chiral_volumes = " refined";
      if (gb.do_geman_mcclure_distances) refined_GMs = " refined";

      std::cout << std::right << std::setprecision(4) << std::fixed;

      gb.print_refinement_flags();
      std::cout << "post-minimize bond   distortion " << std::setw(10) << sum_bond_distortion  << "   " << sqrt(sum_bond_distortion/static_cast<float>(gb.transfer_restraints.bonds.size()))
                << " from " << gb.transfer_restraints.bonds.size() << " bond restraints" << refined_bonds << std::endl;
      std::cout << "post-minimize angle  distortion " << std::setw(10) << sum_angle_distortion << "   " << sqrt(sum_angle_distortion/static_cast<float>(gb.transfer_restraints.angles.size()))
                << " from " << gb.transfer_restraints.angles.size() << " angle restraints" << refined_angles << std::endl;
      if (gb.transfer_restraints.torsions.size() > 0)
         std::cout << "post-minimize torsion distortion " << std::setw(10) << sum_torsion_distortion << "   " << sqrt(sum_torsion_distortion/static_cast<float>(gb.transfer_restraints.torsions.size()))
                   << " from " << gb.transfer_restraints.angles.size() << " angle restraints" << refined_angles << std::endl;
      std::cout << "post-minimize chiral distortion " << std::setw(10) << sum_chiral_volume_distortion << "   "
                << sqrt(sum_chiral_volume_distortion/static_cast<float>(gb.transfer_restraints.chiral_volumes.size())) << " from "
                << std::setw(4) << gb.transfer_restraints.chiral_volumes.size() << " chiral volume restraints" << refined_chiral_volumes << std::endl;
      std::cout << "post-minimize plane  distortion " << std::setw(10) << sum_plane_distortion << "   "
                << sqrt(sum_plane_distortion/static_cast<float>(gb.transfer_restraints.planes.size())) << refined_planes << std::endl;
      std::cout << "post-minimize nbc    distortion " << std::setw(10) << sum_nbc_distortion   << "   " << sum_nbc_distortion/static_cast<float>(gb.transfer_restraints.nbcs.size())
                << refined_nbcs << std::endl;
      std::cout << "post-minimize GM     distortion " << std::setw(10) << sum_geman_mcclure_distortion << "   "
                << sum_geman_mcclure_distortion/static_cast<float>(gb.transfer_restraints.n_geman_mcclures()) << refined_GMs << std::endl;

   // need to clear up nxmap here - do it using cudaFree in the destructor

      cudaFree(d_x);
      delete [] x;
   }

   return atoms_for_return;
}


#endif // __CUDACC__

