
// -*-c++-*-

#ifndef CUDA_RESTRAINTS_HH
#define CUDA_RESTRAINTS_HH

#include <iostream>
#include <sstream>
#include <fstream>
#include <iterator>

#include <stdio.h>
#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n", \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)

#include "cuda-restraints.cuh"

cuda_restraints_transfer_container_t::cuda_restraints_transfer_container_t(const std::string &file_name) {

   std::ifstream f(file_name.c_str());
   if (f) {
      std::string line;
      while (std::getline(f, line)) {
	 if (line.length() > 13) {
	    if (line.substr(0,4) == "BOND")
	       bonds.push_back(cuda_bond_restraint_t(line));
	    if (line.substr(0,5) == "ANGLE")
	       angles.push_back(cuda_angle_restraint_t(line));
	    if (line.substr(0,10) == "CHIRAL_VOL")
	       chiral_volumes.push_back(cuda_chiral_volume_restraint_t(line));
	    if (line.substr(0,10) == "NON_BONDED")
	       nbcs.push_back(cuda_nbc_restraint_t(line));
	    if (line.substr(0,5) == "PLANE")
	       planes.push_back(cuda_host_plane_restraint_t(line));
	    if (line.substr(0,13) == "GEMAN_MCCLURE")
	       geman_mcclures.push_back(cuda_geman_mcclure_restraint_t(line));
	 }
      }
      std::cout << "Parsed " << bonds.size() << " bond restraints" << std::endl;
   } else {
      std::cout << "WARNING:: failed to read restraints file " << file_name << std::endl;
   }
}

float
cuda_restraints_transfer_container_t::sum_bond_distortion(float *x) const {

   printf("Here in cuda_restraints_transfer_container_t::sum_bond_distortion with bonds.size() %d\n", bonds.size());
   float sum = 0.0;
   for (unsigned int i=0; i<bonds.size(); i++) {
      const cuda_bond_restraint_t &r = bonds[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

float
cuda_restraints_transfer_container_t::sum_geman_mcclure_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<geman_mcclures.size(); i++) {
      const cuda_geman_mcclure_restraint_t &r = geman_mcclures[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

float
cuda_restraints_transfer_container_t::sum_angle_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<angles.size(); i++) {
      const cuda_angle_restraint_t &r = angles[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

float
cuda_restraints_transfer_container_t::sum_torsion_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<torsions.size(); i++) {
      const cuda_torsion_restraint_t &r = torsions[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}


float
cuda_restraints_transfer_container_t::sum_plane_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<planes.size(); i++) {
      const cuda_host_plane_restraint_t &r = planes[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

float
cuda_restraints_transfer_container_t::sum_chiral_volume_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<chiral_volumes.size(); i++) {
      const cuda_chiral_volume_restraint_t &r = chiral_volumes[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}



float
cuda_restraints_transfer_container_t::sum_nbc_distortion(float *x) const {

   float sum = 0.0;
   for (unsigned int i=0; i<nbcs.size(); i++) {
      const cuda_nbc_restraint_t &r = nbcs[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

float
cuda_restraints_transfer_container_t::sum_target_pos_distortion(const float *x)  const {

   float sum = 0.0;
   for (unsigned int i=0; i<nbcs.size(); i++) {
      const cuda_target_pos_restraint_t &r = target_positions[i];
      float d = r.distortion(x);
      sum += d;
   }
   return sum;
}

__host__
cuda_device_plane_restraint_t::cuda_device_plane_restraint_t (const cuda_host_plane_restraint_t &rest) {

   n_atom_index_sigma_pairs = rest.atom_indices_with_sigmas.size();

   // reserve some device space for the atom_indices and sigmas:

   // atom indices
   cudaMalloc(&atom_indices, n_atom_index_sigma_pairs * sizeof(int));
   cudaCheckErrors("cudaMalloc cuda_device_plane_restraint_t indices constructor fail");
   int *h_aias_p = new int[n_atom_index_sigma_pairs];
   // std::cout << "--------- transfering " << rest.atom_indices_with_sigmas.size() << " atoms to device "<< std::endl;
   for (unsigned int ii=0; ii<rest.atom_indices_with_sigmas.size(); ii++) {
      if (false)
	 std::cout << "host memory plane atom idx " << ii
		   << " atoms-idx: " << rest.atom_indices_with_sigmas[ii].first
		   << " " << rest.atom_indices_with_sigmas[ii].second << std::endl;
      h_aias_p[ii] = rest.atom_indices_with_sigmas[ii].first;
   }
   cudaMemcpy(atom_indices, h_aias_p, n_atom_index_sigma_pairs * sizeof(int), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMemcpy cuda_device_plane_restraint_t indices constructor fail");
   delete [] h_aias_p;

   // sigmas
   cudaMalloc(&sigmas, n_atom_index_sigma_pairs * sizeof(float));
   cudaCheckErrors("cudaMalloc cuda_device_plane_restraint_t sigmas constructor fail");
   float *h_s_p = new float[n_atom_index_sigma_pairs];
   for (unsigned int ii=0; ii<rest.atom_indices_with_sigmas.size(); ii++)
      h_s_p[ii] = rest.atom_indices_with_sigmas[ii].second;
   cudaMemcpy(sigmas, h_s_p, n_atom_index_sigma_pairs * sizeof(float), cudaMemcpyHostToDevice);
   cudaCheckErrors("cudaMemcpy cuda_device_plane_restraint_t sigmas constructor fail");
   delete [] h_s_p;
}


cuda_restraints_container_t::~cuda_restraints_container_t() {

   // delete

   // bonds
   // nbcs
   // angles
   // chiral_volumes

}

cuda_bond_restraint_t::cuda_bond_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() == 5) {
      if (parts[0] == "BOND") {
         atom_idx_1 = std::stoi(parts[1]);
         atom_idx_2 = std::stoi(parts[2]);
         target_value = std::stof(parts[3]);
         sigma        = std::stof(parts[4]);
      }
   }
   // std::cout << "made bond " << atom_idx_1 << " " << atom_idx_2 << " " << target_value << " " << sigma << "\n";
}

cuda_nbc_restraint_t::cuda_nbc_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() == 4) {
      if (parts[0] == "NON_BONDED") {
         atom_idx_1 = std::stoi(parts[1]);
         atom_idx_2 = std::stoi(parts[2]);
         target_value = std::stof(parts[3]);
         type = LENNARD_JONES;
      }
   }
   // std::cout << "made NBC " << atom_idx_1 << " " << atom_idx_2 << " " << target_value << "\n";
}

cuda_geman_mcclure_restraint_t::cuda_geman_mcclure_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() == 6) {
      if (parts[0] == "GEMAN_MCCLURE") {
         atom_idx_1 = std::stoi(parts[1]);
         atom_idx_2 = std::stoi(parts[2]);
         target_value = std::stof(parts[3]);
         sigma        = std::stof(parts[4]);
         alpha        = std::stof(parts[5]);
      }
   }
   // std::cout << "made GM " << atom_idx_1 << " " << atom_idx_2 << " " << target_value << " " << sigma << " " << alpha << "\n";
}

cuda_angle_restraint_t::cuda_angle_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() == 6) {
      if (parts[0] == "ANGLE") {
         atom_idx_1 = std::stoi(parts[1]);
         atom_idx_2 = std::stoi(parts[2]);
         atom_idx_3 = std::stoi(parts[3]);
         target_value = std::stof(parts[4]);
         sigma        = std::stof(parts[5]);
      }
   }
   // std::cout << "made angle " << atom_idx_1 << " " << atom_idx_2 << " " << atom_idx_3 << " " << target_value << " " << sigma << "\n";
}

cuda_chiral_volume_restraint_t::cuda_chiral_volume_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() == 7) {
      if (parts[0] == "CHIRAL_VOL") {
         atom_idx_c = std::stoi(parts[1]);
         atom_idx_1 = std::stoi(parts[2]);
         atom_idx_2 = std::stoi(parts[3]);
         atom_idx_3 = std::stoi(parts[4]);
         target_chiral_volume = std::stof(parts[5]);
         sigma        = std::stof(parts[6]);
      }
   }
   // std::cout << "made chiral " << atom_idx_c << " " << atom_idx_1 << " " << atom_idx_2 << " " << atom_idx_3 << " " << target_chiral_volume << " " << sigma << "\n";
}

cuda_host_plane_restraint_t::cuda_host_plane_restraint_t(const std::string &s) {

   std::istringstream iss(s);
   std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                  std::istream_iterator<std::string>());

   if (parts.size() > 4) {
      if (parts[0] == "PLANE") {
	 for (std::size_t i=1; i<parts.size()-1; i+=2) {
	    std::pair<unsigned int, float> p(std::stoi(parts[i]), std::stof(parts[i+1]));
	    atom_indices_with_sigmas.push_back(p);
	 }
	 if (false)
	    std::cout << "parsed " << atom_indices_with_sigmas.size() << " atoms in the plane restraint"
		      << std::endl;
      }
   }
}


void
cuda_restraints_transfer_container_t::write(const std::string &file_name) const {

   std::ofstream f(file_name);

   if (f) {
      for(std::size_t i=0; i<bonds.size(); i++)   bonds[i].write(f);
      for(std::size_t i=0; i<angles.size(); i++) angles[i].write(f);
      for(std::size_t i=0; i<chiral_volumes.size(); i++) chiral_volumes[i].write(f);
      for(std::size_t i=0; i<planes.size(); i++) planes[i].write(f);
      for(std::size_t i=0; i<nbcs.size(); i++) nbcs[i].write(f);
      for(std::size_t i=0; i<geman_mcclures.size(); i++) geman_mcclures[i].write(f);
   }

   f.close();

}

void
cuda_bond_restraint_t::write(std::ofstream &f) const {

   std::stringstream ss1, ss2, sst, sss;
   ss1 << atom_idx_1;
   ss2 << atom_idx_2;
   sst << target_value;
   sss << sigma;
   f << "BOND " << ss1.str() << " " << ss2.str() << " " << sst.str() << " " << sss.str() << "\n";
}

void
cuda_nbc_restraint_t::write(std::ofstream &f) const {

   std::stringstream ss1, ss2, sst, sss;
   ss1 << atom_idx_1;
   ss2 << atom_idx_2;
   sst << target_value;
   f << "NON_BONDED " << ss1.str() << " " << ss2.str() << " " << sst.str() << " " << "\n";
}

void
cuda_geman_mcclure_restraint_t::write(std::ofstream &f) const {

   std::stringstream ss1, ss2, sst, sss, ssa;
   ss1 << atom_idx_1;
   ss2 << atom_idx_2;
   sst << target_value;
   sss << sigma;
   ssa << alpha;
   f << "GEMAN_MCCLURE " << ss1.str() << " " << ss2.str() << " " << sst.str() << " " << sss.str() << " " << ssa.str() << "\n";
}

void
cuda_angle_restraint_t::write(std::ofstream &f) const {

   std::stringstream ss1, ss2, ss3, sst, sss;
   ss1 << atom_idx_1;
   ss2 << atom_idx_2;
   ss3 << atom_idx_3;
   sst << target_value;
   sss << sigma;
   f << "ANGLE " << ss1.str() << " " << ss2.str() << " " << ss3.str() << " "
     << sst.str() << " " << sss.str() << "\n";
}

void
cuda_chiral_volume_restraint_t::write(std::ofstream &f) const {

   std::stringstream ssc, ss1, ss2, ss3, sst, sss;
   ssc << atom_idx_c;
   ss1 << atom_idx_1;
   ss2 << atom_idx_2;
   ss3 << atom_idx_3;
   sst << target_chiral_volume;
   sss << sigma;
   f << "CHIRAL_VOL " << ssc.str() << " " << ss1.str() << " " << ss2.str() << " " << ss3.str() << " "
     << sst.str() << " " << sss.str() << "\n";
}

void
cuda_host_plane_restraint_t::write(std::ofstream &f) const {

   std::stringstream sst, sss;
   f << "PLANE ";
   for (std::size_t i=0; i<atom_indices_with_sigmas.size(); i++) {
      const std::pair<unsigned int, float> &p = atom_indices_with_sigmas[i];
      std::stringstream ssi, sss;
      ssi << p.first;
      sss << p.second;
      f << ssi.str() << " " << sss.str() << " " ;
   }
   f << "\n";
}

#endif // CUDA_RESTRAINTS_HH

