
// -*-c++-*-

#ifndef CUDA_RESTRAINT_HH
#define CUDA_RESTRAINT_HH

#include <vector>


namespace cuda_lau {

   // linear algebra utils

   __host__ __device__ float dot(const float3 &d1, const float3 &d2);
   // caller provides result (to be filled)
   __host__ __device__ void cross(const float3 &d1, const float3 &d2, float3 &result); // fill result

}

class cuda_index_value_pair {
   public:
     cuda_index_value_pair(const int &i, const double &v) : idx(i), value(v) {}
     int idx;
     double value;
};


// This needs to contain different types of restraints
// cuda_bond_retraints, cuda_angle_restraints, cuda_nbc_restraints
// so that they can be handled separately.
//
class cuda_bond_restraint_t {
public:
   __device__ __host__ cuda_bond_restraint_t() {
      sigma = 0.02;
      target_value = 4.5;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
   }
   __device__ __host__ cuda_bond_restraint_t(const unsigned int &idx_1, const unsigned int &idx_2, const float &tv, const float &s) :
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      sigma(s),
      target_value(tv) { }
   __host__ cuda_bond_restraint_t(const std::string &s);
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   float sigma;
   float target_value;
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   __device__ void get_gradients(const float *atom_positions, float *d_grad, float *local_gradients) const;
   void write(std::ofstream &f) const;

};

class cuda_geman_mcclure_restraint_t {
public:
   enum type_t { GEMAN_MCCLURE, HARMONIC };
   __device__ __host__ cuda_geman_mcclure_restraint_t() {
      sigma = 0.02;
      target_value = 4.5;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
      alpha = 0.1;
      type_t type;
   }
   __device__ __host__ cuda_geman_mcclure_restraint_t(const unsigned int &idx_1, const unsigned int &idx_2, const float &tv, const float &s) :
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      sigma(s),
      target_value(tv) { type = GEMAN_MCCLURE; alpha = 0.1; }
   __host__ cuda_geman_mcclure_restraint_t(const std::string &s);
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   type_t type;
   float sigma;
   float target_value;
   float alpha;
   __device__ __host__ void set_alpha(const float &f) { alpha = f; }
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;

};

class cuda_angle_restraint_t {
   // __host__ __device__ float dot(const float d1[3], const float d2[3]) const; now a lua
   __host__ __device__ float rad_to_deg(const float &r) const { return  r * 180.0f/M_PI; }
public:
   // target value is in degrees
   __device__ __host__  cuda_angle_restraint_t() {
      sigma = 9.99;
      target_value = 99.99;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
      atom_idx_3 = 2;
   }
   __device__ __host__  cuda_angle_restraint_t(const unsigned int &idx_1, const unsigned int &idx_2, const unsigned int &idx_3,
                                               const float &tv, const float &s) :
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      atom_idx_3(idx_3),
      sigma(s),
      target_value(tv) { }
   __host__ cuda_angle_restraint_t(const std::string &s);
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   unsigned int atom_idx_3;
   float sigma;
   float target_value;
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;

};


class torsion_distortion_gradients_t {
   public:
      torsion_distortion_gradients_t() { }
      __device__ torsion_distortion_gradients_t(const int &P1_idx,
                                     const int &P2_idx,
                                     const int &P3_idx,
                                     const int &P4_idx,
                                     const float *d_x);

      double phi; // the torsion angle, in radians.
      double tan_phi;
      // x
      double dD_dxP1;
      double dD_dxP2;
      double dD_dxP3;
      double dD_dxP4;
      
      // y
      double dD_dyP1;
      double dD_dyP2;
      double dD_dyP3;
      double dD_dyP4;
      
      // z
      double dD_dzP1;
      double dD_dzP2;
      double dD_dzP3;
      double dD_dzP4;
};


class cuda_torsion_restraint_t {
   // __host__ __device__ float dot(const float d1[3], const float d2[3]) const; now a lua
   __host__ __device__ float rad_to_deg(const float &r) const { return  r * 180.0f/M_PI; }
public:
   // target value is in degrees
   __device__ __host__  cuda_torsion_restraint_t() {
      sigma = 9.99;
      target_value = 99.99;
      periodicity = 1;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
      atom_idx_3 = 2;
      atom_idx_4 = 3;
   }

   // tv in degrees
   __device__ __host__  cuda_torsion_restraint_t(const unsigned int &idx_1, const unsigned int &idx_2, const unsigned int &idx_3, const unsigned int &idx_4,
                                                 const float &tv, const float &s, const int &periodicity_in) :
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      atom_idx_3(idx_3),
      atom_idx_4(idx_4),
      sigma(s),
      target_value(tv),
      periodicity(periodicity_in) {
        // convert to Dreiding formulation
        V_jk = 10.0; // just a guess, should depend on sigma.
        n_jk = periodicity;
        phi0_jk = tv;
     }
   __host__ cuda_torsion_restraint_t(const std::string &s);
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   unsigned int atom_idx_3;
   unsigned int atom_idx_4;
   float sigma;
   float target_value;
   int periodicity;
   float V_jk;
   float n_jk;
   float phi0_jk;
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;

};

class cuda_chiral_volume_restraint_t {
public:
   __device__ __host__  cuda_chiral_volume_restraint_t() {
      sigma = 1.00;
      target_chiral_volume = 120.0;
      atom_idx_c = 0;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
      atom_idx_3 = 2;
   }
   __device__ __host__  cuda_chiral_volume_restraint_t(const unsigned int &idx_c,
                                                       const unsigned int &idx_1, const unsigned int &idx_2, const unsigned int &idx_3,
                                                       const float &tv, const float &s) :
      atom_idx_c(idx_c),
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      atom_idx_3(idx_3),
      sigma(s),
      target_chiral_volume(tv) { }
   __host__ cuda_chiral_volume_restraint_t(const std::string &s);
   unsigned int atom_idx_c;
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   unsigned int atom_idx_3;
   float sigma;
   float target_chiral_volume;
   __host__ __device__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;

};

class cuda_plane_distortion_info_t {
public:
   float abcd[4];
   float centre_1[3];
   float centre_2[3];
   float distortion;
   __host__ __device__ cuda_plane_distortion_info_t() { distortion = 0; }
};

__host__ __device__
cuda_plane_distortion_info_t
cuda_distortion_score_plane_internal(const float *d_x, bool calculate_distortion_flag);

class cuda_host_plane_restraint_t {
   __device__ cuda_plane_distortion_info_t cuda_distortion_score_plane_internal(const float *d_x, bool calculate_distortion_flag) const;
public:
   std::vector<std::pair<unsigned int, float> > atom_indices_with_sigmas;
   __host__  cuda_host_plane_restraint_t() { }
   __host__  cuda_host_plane_restraint_t(const const std::vector<std::pair<unsigned int, float> > &atom_indices_with_sigmas_in) :
      atom_indices_with_sigmas(atom_indices_with_sigmas_in) {}
   __host__ cuda_host_plane_restraint_t(const std::string &s);
   __host__ __device__ float distortion(const float *d_x) const;
   void write(std::ofstream &f) const;
};

class cuda_device_plane_restraint_t {
   __device__ cuda_plane_distortion_info_t cuda_distortion_score_plane_internal(const float *d_x, bool calculate_distortion_flag) const;
public:
   
   int n_atom_index_sigma_pairs;
   float *sigmas;
   int *atom_indices;
   __host__ cuda_device_plane_restraint_t (const cuda_host_plane_restraint_t &rest);
   __device__ __host__ cuda_device_plane_restraint_t() { n_atom_index_sigma_pairs = 0;}
   __device__ float distortion(const float *d_x) const;
   __device__ void add_gradients(const float *d_x, float *d_grad) const;

};

class cuda_nbc_restraint_t {
public:
   __device__ __host__  cuda_nbc_restraint_t() {
      target_value = 4.5;
      atom_idx_1 = 0;
      atom_idx_2 = 1;
      type = LENNARD_JONES;
      lj_epsilon = 3.03;
   }
   __device__ __host__  cuda_nbc_restraint_t(const unsigned int &idx_1, const unsigned int &idx_2, const float &tv) :
      atom_idx_1(idx_1),
      atom_idx_2(idx_2),
      target_value(tv) { lj_epsilon = 0.03;
                         sigma = 0.04;  // guess, needs checking
                         type = LENNARD_JONES; }
   __host__ cuda_nbc_restraint_t(const std::string &s);
   enum nbc_type_t { LENNARD_JONES, HARMONIC};
   nbc_type_t type;
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   float target_value;
   float sigma; // for harmonic approx
   float lj_epsilon;
   __device__ void set_lj_epsilon(float l) { lj_epsilon =l; }
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ __host__ float distortion_harmonic_approximation(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   __device__ void add_gradients_harmonic_approximation(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;
};

class cuda_target_pos_restraint_t {
   int atom_idx;
   float3 atom_pull_target_pos;
public:
   __host__ __device__ cuda_target_pos_restraint_t() {}
   __host__ cuda_target_pos_restraint_t(const std::string &s);
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;

};

class cuda_trans_peptide_restraint_t {
public:
   __host__ __device__ cuda_trans_peptide_restraint_t () {}
   __host__ cuda_trans_peptide_restraint_t (const std::string &s);
   unsigned int atom_idx_1;
   unsigned int atom_idx_2;
   unsigned int atom_idx_3;
   unsigned int atom_idx_4;
   __device__ __host__ float distortion(const float *atom_positions) const;
   __device__ void add_gradients(const float *atom_positions, float *d_grad) const;
   void write(std::ofstream &f) const;
};


// this is used on the host, to convert from coot restraints
// and provide easy conversion/transfer to the device
class cuda_restraints_transfer_container_t {
   public:
   cuda_restraints_transfer_container_t() {}
   cuda_restraints_transfer_container_t(const std::string &file_name);
   std::vector<cuda_bond_restraint_t>  bonds;
   std::vector<cuda_nbc_restraint_t>   nbcs;
   std::vector<cuda_angle_restraint_t> angles;
   std::vector<cuda_torsion_restraint_t> torsions;
   std::vector<cuda_chiral_volume_restraint_t> chiral_volumes;
   std::vector<cuda_host_plane_restraint_t> planes;
   std::vector<cuda_geman_mcclure_restraint_t> geman_mcclures;
   std::vector<cuda_target_pos_restraint_t> target_positions;
   std::vector<cuda_trans_peptide_restraint_t> trans_peptides;
   unsigned int n_bonds()  const { return bonds.size(); }
   unsigned int n_angles() const { return angles.size(); }
   unsigned int n_torsions() const { return torsions.size(); }
   unsigned int n_nbcs()   const { return nbcs.size(); }
   unsigned int n_geman_mcclures() const { return geman_mcclures.size(); }
   unsigned int n_target_positions() const { return target_positions.size(); }
   unsigned int n_trans_peptides()   const { return trans_peptides.size(); }

   // shouldn't these be const float *x? Yes, they should.
   float sum_bond_distortion(float *x) const;
   float sum_angle_distortion(float *x) const;
   float sum_torsion_distortion(float *x) const;
   float sum_plane_distortion(float *x) const;
   float sum_nbc_distortion(float *x) const;
   float sum_chiral_volume_distortion(float *x) const;
   float sum_geman_mcclure_distortion(float *x) const;
   float sum_target_pos_distortion(const float *x) const;
 
   void write(const std::string &file_name) const;

};

// this is used on the device
class cuda_restraints_container_t {
 public:
   cuda_restraints_container_t() {
      bonds = NULL;
      nbcs  = NULL;
      angles = NULL;
      torsions = NULL;
      chiral_volumes = NULL;
      planes = NULL;
   }
   ~cuda_restraints_container_t();

   int n_bond_restraints;
   int n_angle_restraints;
   int n_torsion_restraints;
   int n_chiral_volume_restraints;
   int n_nbc_restraints;
   int n_plane_restraints;
   int n_geman_mcclure_restraints;
   cuda_bond_restraint_t  *bonds;
   cuda_nbc_restraint_t   *nbcs;
   cuda_angle_restraint_t *angles;
   cuda_torsion_restraint_t *torsions;
   cuda_chiral_volume_restraint_t *chiral_volumes;
   cuda_device_plane_restraint_t *planes;
   cuda_geman_mcclure_restraint_t *geman_mcclures;
   __device__ float distortion(float *x, bool do_bonds, bool do_angles, bool do_torsions, bool do_chiral_volumes, bool do_planes, bool do_nbcs, bool do_geman_mcclure_distances) const;

};

#endif // CUDA_RESTRAINT_HH
