// -*-c++-*-
#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include <thread>
#include <iterator>
#include <chrono>

#ifndef STANDALONE_INPUTS
#include <clipper/core/coords.h>
#include <clipper/core/nxmap.h>
#include <clipper/core/xmap.h>
#include <clipper/ccp4/ccp4_map_io.h>
#endif // STANDALONE_INPUTS

#include <coot/coot-utils/coot-map-heavy.hh>
#include "nxmap-cuda.cuh"


#ifndef STANDALONE_INPUTS
// on_device arg is optional true
nxmap_cuda::nxmap_cuda(const clipper::NXmap<float> &nxmap, bool on_device) {

   on_host = !on_device;

   clipper::Grid grid_in = nxmap.grid();
   long n_grid_points = grid_in.nu() * grid_in.nv() * grid_in.nw();

   if (true)
      std::cout << "Constructing nxmap_cuda using clipper grid " << grid_in.format() << " " << n_grid_points << std::endl;

   nu_ = grid_in.nu();
   nv_ = grid_in.nv();
   nw_ = grid_in.nw();
   grid = grid_cuda(grid_in.nu(), grid_in.nv(), grid_in.nw());

   std::vector<float> stl_v;
   stl_v.resize(n_grid_points);

   if (true)
      std::cout << "Constructing nxmap_cuda filling grid " << std::endl;

   typedef clipper::NXmap<float>::Map_reference_index NRI;
   unsigned int ic=0;

   auto tp_0 = std::chrono::high_resolution_clock::now();

#if 0
   // this is slow
   for (NRI inx = nxmap.first(); !inx.last(); inx.next()) {
      float dv = nxmap[inx];
      stl_v[ic] = dv;
      ic++;
   }
#endif

   int n_threads = 8;
   std::vector<std::pair<NRI, NRI> > map_ref_start_stops =
      coot::make_map_reference_index_start_stops(nxmap, n_threads);

   std::vector<std::thread> threads;

   auto l = [](std::vector<float> *list_p,
               const clipper::NXmap<float> &nxmap,
               const std::pair<NRI, NRI> &ss) {
      for (NRI inx = ss.first; inx.index() != ss.second.index(); inx.next()) {
         list_p->at(inx.index()) = nxmap[inx];
      }
   };
   for (int i=0; i<n_threads; i++)
      threads.push_back(std::thread(l, &stl_v, std::cref(nxmap), std::cref(map_ref_start_stops[i])));
   for (int i=0; i<n_threads; i++)
      threads[i].join();

   if (!map_ref_start_stops.empty())
      ic = map_ref_start_stops.back().second.index();

   auto tp_1 = std::chrono::high_resolution_clock::now();
   auto d10 = std::chrono::duration_cast<std::chrono::milliseconds>(tp_1 - tp_0).count();
   float n_gigs = 4.0 * static_cast<float>(ic)/static_cast<float>(1024*1024*1024);
   float n_gigs_per_second = 1000.0 * n_gigs / static_cast<float>(d10);
   std::cout << "---------- timings: nxmap stl vector fill: " << ic << " floats, " << n_gigs << " Gb"
             << " in " << d10 << " milliseconds -> " << n_gigs_per_second << " Gb/s\n";



   if (on_device) {
      std::cout << "Constructing nxmap_cuda GPU Malloc of " << n_grid_points << " floats\n";
      if (true)
         std::cout << "Constructing nxmap_cuda getting device memory" << std::endl;
      cudaMalloc(&list_p, n_grid_points * sizeof(float));
      if (true)
         std::cout << "Constructing nxmap_cuda copying data to device" << std::endl;
      cudaMemcpy(list_p, reinterpret_cast<const void *>(stl_v.data()), n_grid_points * sizeof(float), cudaMemcpyHostToDevice);
   } else {
      // for CPU comparisons
      std::cout << "CPU new of " << n_grid_points << " floats\n";
      list_p = static_cast<float *>(new float[n_grid_points]);
      for (unsigned int i=0; i<stl_v.size(); i++)
         list_p[i] = stl_v[i];
   }

   list_size = n_grid_points;

   coord_map_cuda t(nxmap.operator_orth_grid().trn()[0],
                    nxmap.operator_orth_grid().trn()[1],
                    nxmap.operator_orth_grid().trn()[2]);

   mat33_cuda m(nxmap.operator_orth_grid().rot()(0,0),
                nxmap.operator_orth_grid().rot()(0,1),
                nxmap.operator_orth_grid().rot()(0,2),
                nxmap.operator_orth_grid().rot()(1,0),
                nxmap.operator_orth_grid().rot()(1,1),
                nxmap.operator_orth_grid().rot()(1,2),
                nxmap.operator_orth_grid().rot()(2,0),
                nxmap.operator_orth_grid().rot()(2,1),
                nxmap.operator_orth_grid().rot()(2,2));

   rt_orth_grid = rt_cuda(m, t);
   printf("in constructor rt_orth_grid\n");
   rt_orth_grid.print_self();
}
#endif // STANDALONE_INPUTS

nxmap_cuda::nxmap_cuda(const grid_cuda &grid_in, const rt_cuda &rtop_orth_grid_in, bool on_device) {

   on_host = !on_device;
   grid = grid_in;

   long n_grid_points = grid.nu * grid.nv * grid.nw;
   list_size = n_grid_points;

   nu_ = grid.nu;
   nv_ = grid.nv;
   nw_ = grid.nw;

   if (on_device) {
      // add error checking here FIXME
      cudaMalloc(&list_p,   n_grid_points * sizeof(float));
      cudaMemset(list_p, 0, n_grid_points * sizeof(float));
   } else {
     list_p = static_cast<float *>(new float[n_grid_points]);
   }

   std::cout << "on construction of nxmap_cuda() rtop_orth_grid_in:\n" << std::endl;
   rtop_orth_grid_in.print_self();

   rt_orth_grid = rtop_orth_grid_in;
   rt_orth_grid.print_self();

}

// Fill this in with stand-alone data
nxmap_cuda::nxmap_cuda(const std::string &file_name) {

   list_size = 0;
   list_p = 0;
   nu_ = 0;
   nv_ = 0;
   nw_ = 0;

   std::ifstream f(file_name);
   if (f) {
      std::string dim_line, rt_orth_line, rt_grid_line;
      std::getline(f, dim_line);
      // #dim
      if (dim_line.substr(0,4) == "#dim") {
         std::istringstream iss(dim_line);
         std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                        std::istream_iterator<std::string>());
         if (parts.size() == 4) {
            nu_ = std::atoi(parts[1].c_str());
            nv_ = std::atoi(parts[2].c_str());
            nw_ = std::atoi(parts[3].c_str());
            std::cout << "setting grid " << nu_ << " " << nv_ << " " << nw_ << std::endl;
            grid = grid_cuda(nu_, nv_, nw_);
            int n_grid_points = nu_ * nv_ * nw_;
            list_size = n_grid_points;
            std::cout << "map constructor from file CPU allocation (new) of " << n_grid_points << " floats\n";
            list_p = new float[list_size];
         }
      }
      // #orth->grid
      std::getline(f, rt_orth_line);
      if (rt_orth_line.substr(0,11) == "#orth->grid") {
         std::istringstream iss(rt_orth_line);
         std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                        std::istream_iterator<std::string>());
         if (parts.size() == 13) {
            mat33_cuda m(std::atof(parts[1].c_str()), std::atof(parts[2].c_str()), std::atof(parts[3].c_str()),
                         std::atof(parts[4].c_str()), std::atof(parts[5].c_str()), std::atof(parts[6].c_str()),
                         std::atof(parts[7].c_str()), std::atof(parts[8].c_str()), std::atof(parts[9].c_str()));
            coord_orth_cuda trn(std::atof(parts[10].c_str()), std::atof(parts[11].c_str()), std::atof(parts[12].c_str()));
            rt_orth_grid = rt_cuda(m, trn);
	    // rt_orth_grid.print_self();
         }
      }
      // #grid->orth
      std::getline(f, rt_grid_line);
      if (rt_grid_line.substr(0,11) == "#grid->orth") {
         std::istringstream iss(rt_grid_line);
         std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                        std::istream_iterator<std::string>());
         if (parts.size() == 13) {
            mat33_cuda m(std::atof(parts[1].c_str()), std::atof(parts[2].c_str()), std::atof(parts[3].c_str()),
                         std::atof(parts[4].c_str()), std::atof(parts[5].c_str()), std::atof(parts[6].c_str()),
                         std::atof(parts[7].c_str()), std::atof(parts[8].c_str()), std::atof(parts[9].c_str()));
            coord_orth_cuda trn(std::atof(parts[10].c_str()), std::atof(parts[11].c_str()), std::atof(parts[12].c_str()));
            rt_grid_orth = rt_cuda(m, trn);
         }
      }
      std::string line;
      while (std::getline(f, line)) {
         std::istringstream iss(line);
         std::vector<std::string> parts(std::istream_iterator<std::string>{iss},
                                        std::istream_iterator<std::string>());
         if (parts.size() == 4) {
            int u = std::atoi(parts[0].c_str());
            int v = std::atoi(parts[1].c_str());
            int w = std::atoi(parts[2].c_str());
            float f = std::atof(parts[3].c_str()); 
            coord_grid_cuda cg(u,v,w);
            set_data(cg,f);
         }
      }
      f.close();
   }
}

__host__
void
nxmap_cuda::write_map(const std::string &file_name) const {

   if (! on_host) {
     std::cout << "Can only write from host representation" << std::endl;
     return;
   }

   std::ofstream f(file_name);
   if (f) {
      f << "#dim " << nu() << " " << nv() << " " << nw() << "\n";
      f << "#orth->grid ";
      for (unsigned int i=0; i<3; i++)
         for (unsigned int j=0; j<3; j++)
            f << rt_orth_grid.get_rot(i,j) << " ";
      f << rt_orth_grid.get_trn(0) << " ";
      f << rt_orth_grid.get_trn(1) << " ";
      f << rt_orth_grid.get_trn(2) << "\n";
      f << "#grid->orth ";
      for (unsigned int i=0; i<3; i++)
         for (unsigned int j=0; j<3; j++)
            f << rt_grid_orth.get_rot(i,j) << " ";
      f << rt_grid_orth.get_trn(0) << " ";
      f << rt_grid_orth.get_trn(1) << " ";
      f << rt_grid_orth.get_trn(2) << "\n";
      for (unsigned int iu=0; iu<nu(); iu++) {
	 for (unsigned int iv=0; iv<nv(); iv++) {
	    for (unsigned int iw=0; iw<nw(); iw++) {
	       int idx =  iw + iv* nw_ + iu * nw_ * nv_;
	       const float &fl = list_p[idx];
               std::stringstream ssu, ssv, ssw, ssf;
               ssu << iu;
               ssv << iv;
               ssw << iw;
               ssf << fl;
	       f << iu << " " << iv << " " << iw << " " << ssf.str() << "\n";
	    }
	 }
      }
   }
}

void
nxmap_cuda::import(const nxmap_cuda &h_map_in) {

   list_size = h_map_in.size();
   nu_ = h_map_in.nu();
   nv_ = h_map_in.nv();
   nw_ = h_map_in.nw();
   rt_orth_grid = h_map_in.rt_orth_grid;
   rt_grid_orth = h_map_in.rt_grid_orth;
   on_host = false;
   {
      unsigned int n_grid_points = list_size;
      std::cout << "Importing/constructor nxmap_cuda: getting device memory" << std::endl;
      std::cout << "Importing/constructor GPU Malloc of " << n_grid_points << " floats\n";
      cudaMalloc(&list_p, list_size * sizeof(float));
      std::cout << "Importing/constructor nxmap_cuda: copying data to device" << std::endl;
      cudaMemcpy(list_p, h_map_in.list_p, n_grid_points * sizeof(float), cudaMemcpyHostToDevice);
   }

}

