
#include<iostream>
#include<chrono>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <device_launch_parameters.h>
#include <device_functions.h>

#include <CudaLBFGS/error_checking.h>
#include <CudaLBFGS/timer.h>


#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
           fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n",   \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
           fprintf(stderr, "*** FAILED ***\n"); \
        } \
    } while (0)

__device__
int generate_value() {

   int mult = 1;
   if ((blockIdx.x % 2) == 1) {
      mult = -1;
   }
   int prod = threadIdx.x * mult;
   return prod;
}

__global__
void test_kernel_sum_with_partial(int *d_sum_p, int n_slots) {

   int ithread_idx = blockIdx.x * blockDim.x + threadIdx.x;
   int n = 10000000;
   __shared__ int partial_sum;
   partial_sum = 0;

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n; ith += blockDim.x * gridDim.x) {
      // Conclusion: yes you must use atomicAdd for shared memory
      // atomicAdd(&partial_sum, 1);
      atomicAdd(&d_sum_p[0], 1);
   }
   __syncthreads(); // racecheck-report doesn't know that missing this is a problem

   if (ithread_idx == 0) {
      atomicAdd(d_sum_p, partial_sum);
   }
}


__global__
void test_kernel_sum(int *d_sum_p, int n_slots) {

   int n = 1000000;

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < n; ith += blockDim.x * gridDim.x) {
      int islot = ith % n_slots;
      // d_sum_p[islot] += 1;
      atomicAdd(&d_sum_p[islot], 1);
   }

}

__global__
void test_kernel(int *d_result_p) {

   int idx_thread = threadIdx.x;
   int ith = blockIdx.x * blockDim.x + threadIdx.x;
   __shared__ int partial_sum;
   partial_sum = 0;

   // the launch parameters are blockIdx.x and blockDim.x
   // printf("test_kernel() idx_thread %d ith %d blockIdx.x %d blockIdx.x%2 %d \n", idx_thread, ith, blockIdx.x, blockIdx.x%2);

   for (int ith = blockIdx.x * blockDim.x + threadIdx.x; ith < blockDim.x * gridDim.x; ith += blockDim.x * gridDim.x) {
      int prod = generate_value();
      // printf("test_kernel() idx_thread %d ith %d adding %d\n", idx_thread, ith, prod);
      atomicAdd(&partial_sum, prod);
      // printf("test_kernel() idx_thread %d ith %d partial_sum now contains %d\n", idx_thread, ith, partial_sum);
   }

   // printf("test_kernel() idx_thread %d ith %d result %d\n", idx_thread, ith, partial_sum);

   if (idx_thread == 0)
      atomicAdd(d_result_p, partial_sum);
}


void
run_test() {

   bool do_slots = true;
   int n = 10;

   if (do_slots) {
      int *d_result_p;
      int *h_results = new int[n];
      cudaMalloc(&d_result_p, sizeof(int) * n);
      cudaCheckErrors("cudaMalloc d_result_p fail");

      cudaMemset(d_result_p, 0, sizeof(int) * n);
      auto tp_1 = std::chrono::high_resolution_clock::now();
      test_kernel_sum_with_partial<<<1,1024>>>(d_result_p, n);
      cudaCheckErrors("cuda kernel fail test_kernel)");
      cudaDeviceSynchronize();
      cudaCheckErrors("cuda kernel fail sync post test_kernel()");
      auto tp_2 = std::chrono::high_resolution_clock::now();
      auto d21 = std::chrono::duration_cast<std::chrono::microseconds>(tp_2 - tp_1).count();
      std::cout << "info:: d21 " << d21 << " microseconds\n";

      cudaMemcpy(h_results, d_result_p, n * sizeof(int), cudaMemcpyDeviceToHost);
      std::cout << "back in C++, results: " << std::endl;
      for (unsigned int i=0; i<n; i++) {
         std::cout << " " << h_results[i];
      }
      std::cout << std::endl;

    } else {

      int *d_result_p;
      int result = 0;
      cudaMalloc(&d_result_p, sizeof(int));
      cudaCheckErrors("cudaMalloc d_result_p fail");
      cudaMemcpy(d_result_p, &result, sizeof(int), cudaMemcpyHostToDevice);

      auto tp_1 = std::chrono::high_resolution_clock::now();
      // test_kernel<<<3,64>>>(d_result_p);
      test_kernel<<<1,1024>>>(d_result_p);
      cudaCheckErrors("cuda kernel fail test_kernel)");
      cudaDeviceSynchronize();
      cudaCheckErrors("cuda kernel fail sync post test_kernel()");
      auto tp_2 = std::chrono::high_resolution_clock::now();
      auto d21 = std::chrono::duration_cast<std::chrono::microseconds>(tp_2 - tp_1).count();
      std::cout << "info:: d21 " << d21 << " microseconds\n";

      cudaMemcpy(&result, d_result_p, sizeof(int), cudaMemcpyDeviceToHost);
      std::cout << "back in C++, result: " << result << std::endl;

    }

}



int main(int argc, char *argv) {

   run_test();
   return 0;
}

